{  ppmdb -- quick access proper motion database
   Version 20220317
   Copyright (C) 2020-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{$mode objfpc}
unit ppmdb;

interface
uses sysutils;

const ppmdb_chunk_m = 16;       { chunk size in entries is 2^ppdmdb_chunk_m }
      ppmdb_chunks = 16384;     { up to 2^30 entries per tile }


{********* error handling ******************************************************}
(*type errorproc = procedure(l:longint; const s:string);  { l:level (0=error, 1=warning); s:message }
var lasterror : string;                 { last error message }
procedure error(const s:string);        { s:error message }
procedure warning(const s:string);      { s:error message }
procedure stderror(l:longint; const s:string); { s:error message }
procedure InitErrorProc(ep:errorproc);  { initializes new error handler }
procedure RestoreErrorProc;             { restores last error handler } *)

type string2 = string[2];
     string15 = string[15];
     TPPMDB_Colors = array[0..2] of string2;

     CPPMDB_Exception = class(Exception);

     TPPMDB_Header = packed record
{0}     id       : array[0..3] of char;         { must be 'PMDB' }
{4}     dec_n    : word;                        { tile height is 180°/dec_n }
{6}     ra_i     : word;                        { ra index }
{8}     dec_i    : word;                        { dec index }
{10}    mag_n    : word;                        { magnitude step size is 1/mag_n }
{12}    pmma     : word;                        { pm resolution in mas/year }
{14}    epoch    : word;                        { Julian epoch year. This is the reference year for ppm calculation and should be close to observation year in order to reduce rounding errors }
{16}    colors   : array[0..2,0..1] of char;    { color code of magnitude information }
{22}    reserved : array[0..9] of byte;         { reserved, must b 0 }
{32}    mag      : array[0..79] of word;        { magnitude offsets plus 2*mag_n in magnitude steps, i.e. mag[i]=0 means magnitude -2. Total range 80*16=1280 magnitude steps }
{192}   mag_idx  : array[0..79] of longint;     { first index for which a magnitude offset is used }
        end;
    PPPMDB_Header = ^TPPMDB_Header;

    TPPMDB_EEntry = packed record               { encoded entry }
        ra,dec     : word;
        pmra,pmdec : shortint;
        m          : word;
        { Entries are sorted by first magnitude. Magnitude encoding:
          bits 0..3  : 1st magnitude minus header.mag[i] where i is the largest number that satisfies mag_idx[i]<=idx
          bits 4..9  : difference between 2nd magnitude and 1st magnitude or -32 if undefined
          bits 10..15: difference between 3rd magnitude and 1st magnitude or -32 if undefined }
        end;
    PPPMDB_EEntry = ^TPPMDB_EEntry;
    PPPPMDB_EEntry = ^PPPMDB_EEntry;

    TPPMDB_TEntry = packed record               { temporary encoded entry }
        ra,dec     : word;
        pmra,pmdec : shortint;
        m          : array[0..2] of smallint;   { in mag steps, <0 means undefined }
        end;
    PPPMDB_TEntry = ^TPPMDB_TEntry;
    PPPPMDB_TEntry = ^PPPMDB_TEntry;

    TPPMDB_Entry = record                       { decoded entry }
        ra,dec : double;                        { in 2^-16 degrees }
        m      : array[0..2] of smallint;       { in mag steps plus 2*mag_n, <0 means undefined }
        end;

    { helper to calculate tile indexes }
    CPPMDB_Indexer = class
       protected
         vdec_n   : longint;                    { tile height is 180°/dec_n }
       public
         constructor create(dec_n:longint);     { init using number of tiles in dec direction }
         constructor create(const dbd:ansistring);      { init using database (dbd/00_00.pdb is used to read parameters) }
         function ra_n(dec_i:longint):longint;
         procedure radec_i(ra,dec:double; var ra_i,dec_i:longint);
         function radec_i(ra,dec:double):string15;
         property dec_n : longint read vdec_n;          { tile height is 180°/dec_n }
    end;

    { represents a tile }
    CPPMDB_Tile = class
       private
         header      : TPPMDB_Header;
         chunks      : array[0..ppmdb_chunks] of PPPMDB_EEntry;
         vsize       : longint;                                 { number of entries }
         vcolors     : TPPMDB_Colors;
         prev_mi     : longint;                                 { previous mag table index for faster sequential access }
         ra_f,dec_f,pm_f  : double;                             { conversion factors}
         ra_o,dec_o       : double;                             { offsets }
         procedure reset;                                       { reset chunks }
       public
         constructor create(const fn:ansistring);               { reads tile from file }
         destructor destroy; override;
         procedure write(const fn:ansistring);                  { writes tile to a file }
         function test(var msg:shortstring):boolean;            { tests tile }
         property dec_n  : word read header.dec_n;              { tile height is 180°/dec_n }
         property mag_n  : word read header.mag_n;              { magnitude step size is 1/mag_n }
         property colors : TPPMDB_Colors read vcolors;          { color code of magnitude information }
         property size   : longint read vsize;
         property ra_i   : word read header.ra_i;               { ra index }
         property dec_i  : word read header.dec_i;              { dec index }
       public
         didx  : longint;
         dbuf  : TPPMDB_Entry;
         function decode(year,idx:longint; var buf:TPPMDB_Entry):boolean;       { random access decode }
         function decode(year:longint):boolean;                 { sequential decode using didx (will be increased) as index and dbuf for output. If invalid, magnitudes of dbuf will be set to -1  }
    end;

    { represents a tile buffer which is used for creation }
    CPPMDB_Creator = class
       private
         header   : TPPMDB_Header;
         chunks   : array[0..ppmdb_chunks] of PPPMDB_TEntry;
         hist     : ^longint;
         hist_n   : longint;
         vsize    : longint;                                    { number of entries }
         vcolors  : TPPMDB_Colors;
         procedure reset;                                       { reset chunks }
       public
         constructor create(epoch,dec_n,ra_i,dec_i,mag_n,pmma,maxmag:longint; c1,c2,c3:string2);  { maxmag: maximum magnitude in encoded form, i.e. (mag+2)*mag_n }
         destructor destroy; override;
         procedure append(te:PPPMDB_TEntry);                    { unsorted append  }
         procedure write(const fn:ansistring);                  { sorts the buffer and writes tile to a file }
         property dec_n  : word read header.dec_n;              { tile height is 180°/dec_n }
         property mag_n  : word read header.mag_n;              { magnitude step size is 1/mag_n }
         property colors : TPPMDB_Colors read vcolors;          { color code of magnitude information }
         property size   : longint read vsize;
    end;

procedure ppmdb_normalize(var ra,dec:double);           { dec will between -90° and 90°, ra will be between 0° and 360° }
function ppmdb_fileBase(ra_i,dec_i:longint):string15;


{*******************************************************************************}
{*******************************************************************************}
{********* implementation ******************************************************}
{*******************************************************************************}
{*******************************************************************************}
implementation
uses baseunix,errors;
const ppmdb_chunk_size    = 1 shl ppmdb_chunk_m;
      ppmdb_chunk_mask    = ppmdb_chunk_size - 1;

{********* ppmdb_normalizeDec **************************************************}
procedure ppmdb_normalize(var ra,dec:double);
begin
    while dec>90 do
    begin
        dec:=180-dec;
        ra+=180;
    end;
    while dec<-90 do
    begin
        dec:=-180-dec;
        ra+=180;
    end;
    ra-=round(ra/360)*360;
    if ra<0 then ra+=360;
end;

{********* ppmdb_fileBase ******************************************************}
function ppmdb_fileBase(ra_i,dec_i:longint):string15;
var sra,sdec   :  string[7];
begin
    str(ra_i,sra);      if length(sra)<2 then sra:='0'+sra;
    str(dec_i,sdec);    if length(sdec)<2 then sdec:='0'+sdec;
    result:=sdec+'_'+sra;
end;

{********* fpread2 *************************************************************}
{ tries to read at least len bytes are read
  returns number of bytes read or -1 on error }
function fpread2(fd:longint; buf:pchar; len:longint):longint;
var res : longint;
begin
    res:=1;
    result:=0;
    while (res>0) and (len>result) do
    begin
        res:=fpread(fd, buf, len-result);
        if res>0 then
        begin
            buf+=res;
            result+=res;
        end;
    end;
    if res<0 then result:=res;
end;

{********* ppmdb_read **********************************************************}
{ returns empty string on success, otherwise an error message
  fn    : File name
  h     : header
  chunks: chunk buffers, can be null if only the header shall be read.
  maxlen: maximum number number of chunks to be read, can b 0
  size  : returns number of entries read
}
function ppmdb_read(const fn:ansistring; var h:TPPMDB_Header; chunks:PPPPMDB_EEntry; maxlen:longint; var size:longint):ansistring;
var fd,i,j      : longint;
begin
    result:='';
    fd:=fpopen(fn,o_rdOnly);
    if fd<0 then
    begin
        result:='Error opening '+fn+': '+StrError(fpGetErrno);
        exit;
    end;
    if fpread2(fd,pchar(@h),sizeof(TPPMDB_Header))<sizeof(TPPMDB_Header) then
    begin
        result:='Error reading '+fn+': '+StrError(fpGetErrno);
        exit;
    end;
    if (h.id<>'PMDB' ) then
    begin
        result:='Error reading '+fn+': Invalid magic number / wrong file format';
        exit;
    end;
    size:=0;
    if (chunks<>nil) and (maxlen>0) then
    begin
        i:=0;
        repeat
            getmem(chunks[i], sizeof(TPPMDB_EEntry)*ppmdb_chunk_size);
            j:=fpread2(fd, pchar(chunks[i]), 8*ppmdb_chunk_size) >> 3;
            size+=j;
            if j=0 then freemem(chunks[i],8*ppmdb_chunk_size)
            else i+=1;
        until (j<ppmdb_chunk_size) or (i>=maxlen);
    end;
    fpclose(fd);
end;


{********* ppmdb_write *********************************************************}
{ Writes tile to file and returns empty string on success, otherwise an error message
  fn    : File name
  h     : header
  chunks: chunk buffers, can be null if only the header shall be written
  chunk_n : if null ppmdb_chunk_size is
  size  : number of entries to write
}
function ppmdb_write(const fn:ansistring; h:PPPMDB_Header; chunks:PPPPMDB_EEntry; chunk_n:plongint; size:longint):ansistring;
var fd,i,j,k  : longint;
begin
    result:='';
    fd:=fpopen(fn,o_wrOnly or o_Creat );
    if fd<0 then
    begin
        result:='Error opening '+fn+': '+StrError(fpGetErrno);
        exit;
    end;
    if fpwrite(fd,pchar(h),sizeof(TPPMDB_Header))<sizeof(TPPMDB_Header) then result:='Error writing '+fn+': '+StrError(fpGetErrno);
    if chunks=nil then size:=0;
    i:=0;
    while (size>0) and (result='') do
    begin
        j:=size;
        if chunk_n<>nil then k:=chunk_n[i]
        else k:=ppmdb_chunk_size;
        if j>k then j:=k;
        if ( j>0 ) then
        begin
            if fpwrite(fd,pchar(chunks[i]),j*sizeof(TPPMDB_EEntry))<j*sizeof(TPPMDB_EEntry) then result:='Error writing '+fn+': '+StrError(fpGetErrno);
            size-=j;
        end;
        i+=1;
    end;
    fpclose(fd);
end;


{*******************************************************************************}
{********* CPPMDB_Indexer ******************************************************}
{*******************************************************************************}
{********* create **************************************************************}
constructor CPPMDB_Indexer.create(dec_n:longint);
begin
    vdec_n:=dec_n;
    if (dec_n<1) or (dec_n>127) then raise CPPMDB_Exception.create('Number of tiles in dec direction must be between 1 and 127');
end;

{********* create **************************************************************}
constructor CPPMDB_Indexer.create(const dbd:ansistring);        { reads tile from file }
var header      : TPPMDB_Header;
    i           : longint;
    res         : ansistring;
begin
    header.dec_n:=0;    // makes the compiler happy
    res:=ppmdb_read(dbd+pathDelim+ppmdb_fileBase(0,0)+'.pdb', header,nil,0,i);
    if res<>'' then raise CPPMDB_Exception.create(res);
    vdec_n:=header.dec_n;
    if (dec_n<1) or (dec_n>127) then raise CPPMDB_Exception.create('Number of tiles in dec direction must be between 1 and 127');
end;

{********* ra_n ****************************************************************}
{ get number of of tiles in ra direction }
function CPPMDB_Indexer.ra_n(dec_i:longint):longint;
var r : double;
begin
    if dec_i<0 then dec_i:=0;
    if dec_i>=dec_n then dec_i:=dec_n-1;
    if 2*(dec_i+1)<dec_n then r:=sin((dec_i+1)/dec_n*pi)
    else if 2*dec_i>dec_n then r:=sin(dec_i/dec_n*pi)
    else r:=1;
    result:=round(dec_n*2*r+0.499999);
end;

{********* radec_i *************************************************************}
{ get ra and dec indexes; arguments in degree }
procedure CPPMDB_Indexer.radec_i(ra,dec:double; var ra_i,dec_i:longint);
var _ra_n : longint;
begin
    ppmdb_normalize(ra,dec);
    dec_i:=trunc((90-dec)/180.0*dec_n) mod dec_n;
    _ra_n:=ra_n(dec_i);
    ra_i:=trunc(ra/360.0*_ra_n) mod _ra_n;
    if ra_i<0 then ra_i+=_ra_n;
end;

{********* radec_i *************************************************************}
function CPPMDB_Indexer.radec_i(ra,dec:double):string15;
var ra_i,dec_i : longint;
begin
    radec_i(ra,dec, ra_i,dec_i);
    result:=ppmdb_fileBase(ra_i,dec_i);
end;


{*******************************************************************************}
{********* CPPMDB_Tile *********************************************************}
{*******************************************************************************}
{********* create **************************************************************}
constructor CPPMDB_Tile.create(const fn:ansistring);    { reads tile from file }
var i,j  : longint;
    res  : ansistring;
    indexer : CPPMDB_Indexer;
begin
    res:=ppmdb_read(fn, header,@chunks,ppmdb_chunks, vsize);
    if res<>'' then raise CPPMDB_Exception.create(res);

    for i:=0 to 2 do
    begin
        vcolors[i]:='';
        for j:=0 to 1 do
            if header.colors[i,j]<>#0 then vcolors[i]+=header.colors[i,j];
    end;
    prev_mi:=0;
    didx:=0;

    indexer:=CPPMDB_Indexer.create(header.dec_n);
    ra_f :=360/indexer.ra_n(dec_i);
    ra_o:=ra_f*ra_i;
    ra_f/=65536.0;

    dec_f:=180/header.dec_n;
    dec_o:=90-dec_f*dec_i;
    dec_f/=-65536.0;

    pm_f :=header.pmma/(1000*3600);
    indexer.destroy;

{    for i:=0 to 79 do
        writeln(header.mag_idx[i],': ',header.mag[i]); }
end;

{********* destroy *************************************************************}
destructor CPPMDB_Tile.destroy;
begin
    reset;
    inherited destroy;
end;

{********* reset ***************************************************************}
procedure CPPMDB_Tile.reset;
var i : longint;
begin
    if size<1 then exit;
    for i:=0 to (size-1) shr ppmdb_chunk_m do;
        freemem(chunks[i], sizeof(TPPMDB_EEntry)*ppmdb_chunk_size);
    vsize:=0;
end;

{********* write ***************************************************************}
procedure CPPMDB_Tile.write(const fn:ansistring);
var res  : ansistring;
begin
    res:=ppmdb_write(fn, @header, @chunks, nil, size);
    if res<>'' then raise CPPMDB_Exception.create(res);
end;

{********* decode **************************************************************}
{ random access decode }
function CPPMDB_Tile.decode(year,idx:longint; var buf:TPPMDB_Entry):boolean;
const deg2rad = pi/180;
var ee     : PPPMDB_EEntry;
    i,j,mi : longint;
    r      : double;
    w      : word;

begin
    year-=header.epoch;
    if (idx<0) or (idx>=vsize) then
    begin
        result:=false;
        exit;
    end;

    result:=true;
    ee:=@chunks[idx shr ppmdb_chunk_m][idx and ppmdb_chunk_mask];
    { find mag table entry }
    mi:=prev_mi;
    while (mi<79) and (header.mag_idx[mi+1]<=idx) do
        mi+=1;
    while (mi>0) and (header.mag_idx[mi]>idx) do
        mi-=1;
    prev_mi:=mi;
    { decode }
    w:=ee^.m;
    buf.m[0]:=(ee^.m and 15) + header.mag[mi];
    w:=w shr 4;
    for j:=1 to 2 do
    begin
        i:=w and 63;
        if i<32 then buf.m[j]:=buf.m[0]+i
        else if i>32 then buf.m[j]:=buf.m[0]+i-64
        else buf.m[j]:=-1;
        w:=w shr 6;
    end;
    buf.dec:=ee^.dec*dec_f + dec_o + ee^.pmdec*year*pm_f;
    r:=cos(buf.dec*deg2rad);
    if r<1e-10 then r:=1e-10;
    buf.ra :=ee^.ra*ra_f + ra_o + ee^.pmra*year*pm_f/r;
    ppmdb_normalize(buf.ra,buf.dec);
end;

{ sequential decode using didx (will be increased) as index and dbuf for output }
function CPPMDB_Tile.decode(year:longint):boolean;
begin
    result:=decode(year,didx,dbuf);
    if result then didx+=1
    else begin
        dbuf.m[0]:=-1;
        dbuf.m[1]:=-1;
        dbuf.m[2]:=-1;
        dbuf.ra:=0;
        dbuf.dec:=0;
    end;
end;

{********* test ****************************************************************}
{ returns true if valid
  msg : info string if valid or error }
function CPPMDB_Tile.test(var msg:shortstring):boolean;
const year = 2020;
var i,m  : longint;
begin
    msg:='';
    didx:=0;
    m:=0;
    for i:=0 to size-1 do
    begin
        decode(year);
//      writeln(i,'  ',dbuf.m[0],'   ',dbuf.m[1],'   ',dbuf.m[2]);
        if dbuf.m[0]<m then
        begin
            writeln(didx-1,'/',size,': ',header.mag_idx[prev_mi],'  ',header.mag[prev_mi],' -  ',dbuf.m[0],'   ',m);
            msg:='Sources are not sorted or magnitude table is corrupt.';
            result:=false;
            exit;
        end;
        m:=dbuf.m[0];
    end;
    didx:=0;
    decode(year);
    msg:=format('%7d sources: ( mag0=%-.2f',[size,dbuf.m[0]/mag_n-2]);
    for i:=1 to 2 do
        if dbuf.m[i]>=0 then msg+=format(', mag%d=%-.2f',[i,dbuf.m[i]/mag_n-2]);

    didx:=size-1;
    decode(year);
    msg+=format(' ) ... ( mag0=%-.2f',[dbuf.m[0]/mag_n-2]);
    for i:=1 to 2 do
        if dbuf.m[i]>=0 then msg+=format(', mag%d=%-.2f',[i,dbuf.m[i]/mag_n-2]);
    msg+=' )';
    result:=true;
end;

{*******************************************************************************}
{********* CPPMDB_Creator ******************************************************}
{*******************************************************************************}
{********* create **************************************************************}
constructor CPPMDB_Creator.create(epoch,dec_n,ra_i,dec_i,mag_n,pmma,maxmag:longint; c1,c2,c3:string2);
var i   : longint;
begin
    header.id:='PMDB';
    header.dec_n:=dec_n;
    header.ra_i:=ra_i;
    header.dec_i:=dec_i;
    header.mag_n:=mag_n;
    header.pmma:=pmma;
    header.epoch:=epoch;
    fillbyte(header.reserved,sizeof(header.reserved),0);
    vcolors[0]:=c1; vcolors[1]:=c2; vcolors[2]:=c3;
    fillbyte(header.colors,6,0);
    for i:=0 to 2 do
    begin
        if length(vcolors[i])>0 then header.colors[i,0]:=vcolors[i][1];
        if length(vcolors[i])>1 then header.colors[i,1]:=vcolors[i][2];
    end;
    vsize:=0;
    hist_n:=maxmag+1;
    getmem(hist,hist_n*sizeof(longint));
    fillbyte(hist^,hist_n*sizeof(longint),0);
end;

{********* destroy *************************************************************}
destructor CPPMDB_Creator.destroy;
begin
    reset;
    getmem(hist,hist_n*sizeof(longint));
    inherited destroy;
end;

{********* reset **************************************************************}
procedure CPPMDB_Creator.reset;
var i  : longint;
begin
    if vsize<1 then exit;
    for i:=0 to (size-1) shr ppmdb_chunk_m do;
        freemem(chunks[i], sizeof(TPPMDB_TEntry)*ppmdb_chunk_size);
    vsize:=0;
    fillbyte(hist^,hist_n*sizeof(longint),0);
end;

{********* append **************************************************************}
procedure CPPMDB_Creator.append(te:PPPMDB_TEntry);
var i,j    : longint;
begin
    i:=te^.m[0];
    if (i<0) or (i>=hist_n) then exit;
    hist[i]+=1;
    i:=vsize shr ppmdb_chunk_m;
    j:=vsize and ppmdb_chunk_mask;
    if j=0 then getmem(chunks[i], sizeof(TPPMDB_TEntry)*ppmdb_chunk_size);
    chunks[i][j]:=te^;
    vsize+=1;
end;

{********* convert *************************************************************}
procedure CPPMDB_Creator.write(const fn:ansistring);
var  sortbuf    : PPPPMDB_EEntry;
     res        : ansistring;
     magoffs    : ^smallint;
     m0,m1,md,m : integer;
     w          : word;
     i,j,k,l    : longint;
     ee         : PPPMDB_EEntry;
     te         : PPPMDB_TEntry;
begin
    if vsize<1 then exit;
    { magnitude indexes }
    getmem(magoffs,hist_n*sizeof(smallint));
    fillbyte(magoffs^,hist_n*sizeof(smallint),0);
    m1:=hist_n-1;
    while hist[m1]=0 do
        m1-=1;
    m0:=0;
    k:=0;
    for i:=0 to 79 do
    begin
        while (m0<=m1) and (hist[m0]=0) do
            m0+=1;
        md:=0;
        j:=hist[m0];
        while (md<16) and (m0+md<m1) and  // maximum criterion
            ( (m0+md<m1+1-(79-i)*16) or   // at least (m1-m0)/(80-i) mag steps
               ((j*(80-i)<=(vsize-k)) ) ) do      // try to split into equal portions
        begin
            md+=1;
            j+=hist[m0+md];
        end;
        j-=hist[m0+md];
        header.mag[i]:=m0;
        header.mag_idx[i]:=k;
//      writeln(header.mag_idx[i],': ',header.mag[i],' + ',md,' (',j,'  ',(vsize-k)/(80-i):0:2,')');

        for l:=0 to md-1 do
            magoffs[m0+l]:=m0;

        m0+=md;
        k+=j;
    end;
//    writeln(vsize,': ',m1);
    for l:=m0 to m1 do
        magoffs[l]:=m0-md;

    { sort and convert data }
    getmem(sortbuf,sizeof(PPPMDB_EEntry)*hist_n);
    for i:=0 to hist_n-1 do
        if hist[i]>0 then getmem(sortbuf[i],hist[i]*sizeof(TPPMDB_EEntry));
    fillbyte(hist^,hist_n*sizeof(longint),0);  { use hist as data pointer }

    for i:=0 to size-1 do
    begin
        te:=@chunks[i shr ppmdb_chunk_m][i and ppmdb_chunk_mask];
        m0:=te^.m[0];
        ee:=@sortbuf[m0][hist[m0]];
        hist[m0]+=1;
        md:=m0-magoffs[m0];
        if (md<0) or (md>15) then raise CPPMDB_Exception.create(format('CPMDB_TileCreate.convert: internal error: corrupt magnitude index: i=%d md=%d mag=%d mag_offs=%d',[i,md,m0,magoffs[m0]]));
//      if i=0 then writeln(format('i=%d md=%d mag=%d mag_offs=%d',[i,md,m0,magoffs[m0]]));
        w:=0;
        for l:=0 to 1 do
        begin
            m:=te^.m[2-l];
            if m<0 then m:=-32
            else begin
                m-=m0;
                if m<-31 then m:=-31
                else if m>31 then m:=31;
            end;
            w:=(w shl 6) or (m and 63);
        end;
        w:=(w shl 4) or (md and 15);
        ee^.ra:=te^.ra;
        ee^.dec:=te^.dec;
        ee^.pmra:=te^.pmra;
        ee^.pmdec:=te^.pmdec;
        ee^.m:=w;
    end;

    { write data }
    res:=ppmdb_write(fn, @header, sortbuf, hist, size);
    if res<>'' then raise CPPMDB_Exception.create(res);

    { dispose buffers }
    freemem(magoffs,hist_n*sizeof(smallint));
    for i:=0 to hist_n-1 do
        if hist[i]>0 then freemem(sortbuf[i],hist[i]*sizeof(TPPMDB_EEntry));
    freemem(sortbuf,sizeof(PPPMDB_EEntry)*hist_n);
end;

end.
