{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name bayer}
{@description Convert Bayer images}
{@longdescription Convert Bayer pattern images.
The Bayer pattern can be specified using -p or via image property "BAYERPAT".
If input is a multi-color image it is converted to a Bayer pattern image first,
otherwise input is treated as Bayer pattern image. }

{$I directives.pas}
uses simage,getparam;
var par                 : tparams;
    cc                  : CSImg_Config;
    ifn,ofn,v,c         : ansistring;
    ps                  : shortstring;
    fi,fj               : TSImg_FPImage;
    bi                  : TSImg_BayerInfo;
    im,il,id,ii,iw      : boolean;
    sc,cf,bs            : boolean;
    split               : boolean;
    black,r             : double;
    i,j,ci              : longint;
    mi,miold            : CSImg_MetaInfo;

begin
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        ifn:='';        defparam('<infile>','Input image (default: stdin)',ifn);
        ofn:='';        defparam('<outfile>','Output image (default: stdout). If filename contains `%c'' output is split into 1 image per color where %c is replaced by the color number starting from 1.',ofn);
                        defparam('-h','Help',helpf);
        ps:='12/34';    defparam('-p','Set Bayer pattern: e.g. 12/23 or RG/GB (both are the same); or 12/43 or RG/WB (both are the same); default: 12/34', ps );

        black:=0;       defparam('-0','Black value (default: 0)',black);

        im:=false;      defparam('-im','Median interpolation',im);
        il:=false;      defparam('-il','Linear interpolation',il);
        ii:=false;      defparam('-iq','25% quantile interpolation',ii);
        id:=false;      defparam('-id','Down-scaling',id);
        iw:=false;      defparam('-iw','Weight masking',iw);

        cf:=false;      defparam('-cf','Flatfield correction: 95% quantile of each color becomes one', cf);
        bi.cf[0]:=1;    defparam('-c1','Correction factor for color 1',bi.cf[0]);
        bi.cf[1]:=1;    defparam('-c2','Correction factor for color 2',bi.cf[1]);
        bi.cf[2]:=1;    defparam('-c3','Correction factor for color 3',bi.cf[2]);
        bi.cf[3]:=1;    defparam('-c4','Correction factor for color 4',bi.cf[3]);

        sc:=false;      defparam('-sc','Saturation compensation (interpolation only)',sc);

        ci:=-1;         defparam('-ci','Interpolate only this color',ci);

                        defparam('-o','Destination image',ofn); deffmt('-o','<outfile>');
        run;

        cc:=CSImg_Config.create(SImg_Config_Default);
        cc.debug('bayer: src1='+ifn+' dest='+ofn+' ps='+ps
            +' im='+boolean2str(im)+' il='+boolean2str(il)+' id='+boolean2str(id)+' ii='+boolean2str(ii)+' iw='+boolean2str(ii)
            +' cf='+condstring(cf, 'flat', '('+real2str(bi.cf[0])+', '+real2str(bi.cf[1])+', '+real2str(bi.cf[2])+', '+real2str(bi.cf[3])+')')
            +' sc='+boolean2str(sc)+' ci='+int2str(ci));


        bs:=false;
        for i:=0 to 3 do
            bs:=bs or (abs(bi.cf[i]-1)>1e-9);

        i:=0;
        if im then i+=1;
        if il then i+=1;
        if id then i+=1;
        if ii then i+=1;
        if iw then i+=1;
        if i>1 then cc.error('bayer: Can''t be defined together: -im, -il, -id, -ii, -iw');
        split:=(i>0) and (pos('%c',ofn)>0);

        if sc and (i=0) then
        begin
            cc.warning('Saturation compensation (option -sc) only has an effect with interpolation');
            sc:=false;
        end;

        mi:=CSImg_MetaInfo.create;
        cc.Info('bayer: Reading input image '+ifn);
        SImg_ReadImageFile(cc,ifn,fi,mi);

        if (not paramdef('-p')) and mi.get('BAYERPAT',v) then ps:=trimquotes(v);

        if i>0 then mi.delete('BAYERPAT')
        else mi.add('BAYERPAT',''''+ps+'''','Bayer color pattern, e.g. ''RG/GB'' or ''01/23''');

        bi.pattern:=SImg_BayerParsePattern(cc,ps);
        if not SImg_BayerCheckInfo(bi) then cc.error('bayer: Invalid Bayer pattern');

        if fi.nc>1 then
        begin
            cc.Info('bayer: '+int2str(fi.nc) + ' color image found, converting to '''+ps+''' Bayer pattern image');
            SImg_ToBayer(cc,fi,fi,bi);
        end;

        if abs(black)>1e-20 then
        begin
            if mi.get('DATAMIN',r,@c) then mi.add('DATAMIN',real2str(r-black,15),c);
            if mi.get('DATAMAX',r,@c) then mi.add('DATAMAX',real2str(r-black,15),c);
            mi.delete('EGAIN');
            mi.delete('BUNIT');
            for i:=0 to fi.size-1 do
                fi.pt[i]-=black;
        end;

        miold:=CSImg_MetaInfo.create(mi);

        if bs then
        begin
            mi.delete('DATAMIN');
            mi.delete('DATAMAX');
            mi.delete('EGAIN');
            mi.delete('BUNIT');
            mi.delete('RDNOISE');
        end;

        if cf then
        begin
            cc.Info('bayer: Color flattening');
            SImg_BayerFlat(cc,fi,bi);
        end;

        if id or il or im or ii or iw then cc.Info('bayer: Color interpolation using ' +int2str(cc.threads) + ' threads');
        if id then SImg_BayerDownscale(cc,bi,fi,fi)
        else if iw then SImg_BayerInterpolation(cc,bi,fi,fi,-2, ci)
        else if ii then SImg_BayerInterpolation(cc,bi,fi,fi,-1, ci)
        else if il then SImg_BayerInterpolation(cc,bi,fi,fi, 0, ci)
        else if im then SImg_BayerInterpolation(cc,bi,fi,fi, 1, ci)
        else if bs then SImg_BayerScale(cc,fi,bi);

        if sc then
        begin
            cc.Info('bayer: Saturation compensation');
            SImg_SaturationCompensation(cc, fi);
        end;

        if split then
        begin
            j:=pos('%c',ofn);
            delete(ofn,j,1);
            fj.pt:=nil;
            for i:=0 to fi.nc-1 do
            begin
                ofn[j]:=char(49+i);
                if miold.get('DATAMIN',r,@c) then mi.add('DATAMIN',real2str(r*abs(bi.cf[i]),15), c);
                if miold.get('DATAMAX',r,@c) then mi.add('DATAMAX',real2str(r*abs(bi.cf[i]),15), c);
                if miold.get('RDNOISE',r,@c) then mi.add('RDNOISE',real2str(r*abs(bi.cf[i])), c);
                if miold.get('EGAIN',r,@c) and (abs(bi.cf[i])>1e-20) then mi.add('EGAIN',real2str(r/abs(bi.cf[i])), c);
                cc.info('bayer: Writing output image '+ofn);
                SImg_Cut(cc,fi,fj, 0,0, fi.nx-1,fi.ny-1, i,i, 0, 0);
                SImg_WriteImageFile(cc,ofn,fj,mi);
            end;
            SImg_Free(cc,fj);
        end
        else begin
            cc.info('bayer: Writing output image '+ofn);
            SImg_WriteImageFile(cc,ofn,fi,mi);
        end;

        SImg_Free(cc,fi);
        done;
    end;

    cc.destroy;
end.
