{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name expsources}
{@description Export sources either as fits table (xy list), as plot, or as csv with column headers in first row. }

{$I directives.pas}
uses simage,getparam,sysutils;
var par        : tparams;
    cc         : CSImg_Config;
    ifn,ofn    : ansistring;
    fwhm,sc    : double;
    ft         : boolean;
    mi         : CSImg_MetaInfo;
    fi         : TSImg_FPImage;
    sl,ri      : TSImg_RDImage;
    out        : CSImg_FileOutput;
    tout       : text;
    dd         : ^double;
    xmin,xmax  : double;
    ymin,ymax  : double;
    y          : longint;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        ifn:='';         defparam('<infile>','Source list (default: stdin)',ifn);
        ofn:='';         defparam('<outfile>','Output file (default: stdout)',ofn);
        fwhm:=-1;        defparam('-p','Export as plot with that fwhm. (Default: tsv table)',fwhm);
        ft:=false;       defparam('-t','Export fits table. (Default: tst table)',ft);
        sc:=1;           defparam('-s','Pixel scale / downscaling factor for plot. Default: 1',sc);
                         defparam('-o','Output source list',ofn);    deffmt('-o','<outfile>');
                         defparam('-h','Help',helpf);
        run;
        done;
    end;

    cc.debug('gexpources: in='+ifn+' out='+ofn+' ft='+boolean2str(ft)+' fwhm='+real2str(fwhm)+' sc='+real2str(sc));

    { input image }
    sl.pt:=nil;
    mi:=CSImg_MetaInfo.create();
    cc.Info('expsources: Reading source list '+ifn);
    SImg_ReadImageFile(cc,ifn,sl,mi);
    if (sl.nx<5) or (sl.fmt<>SImg_RawDouble) or (sl.nc<>1) then cc.Error('expsources: Invalid format of source list');
    if not mi.get('IBNDXMIN', xmin) then cc.error('SImg_PlotSources: Attribute not found: "IBNDXMIN"');
    if not mi.get('IBNDXMAX', xmax) then cc.error('SImg_PlotSources: Attribute not found: "IBNDXMAX"');
    if not mi.get('IBNDYMIN', ymin) then cc.error('SImg_PlotSources: Attribute not found: "IBNDYMIN"');
    if not mi.get('IBNDYMAX', ymax) then cc.error('SImg_PlotSources: Attribute not found: "IBNDYMAX"');

    { TSV table }
    if (not ft) and ( fwhm < 1e-3 ) then
    begin
        if (ofn='') or (ofn='-') then
        begin
            flush(output);
            tout:=output;
        end else
        begin
            {$I-}
            assign(tout,ofn);
            rewrite(tout);
            if ioresult<>0 then cc.error('Error opening '+ofn);
            {$I+}
        end;
        writeln(tout,'# X       Y               Flux    Background      Weight');
        for y:=0 to sl.ny-1 do
        begin
            dd:=sl.ptd + y*sl.nx;
            writeln(tout, dd[0]-xmin:0:1,'      ', dd[1]-ymin:0:1,'     ', dd[2],'      ', dd[3],'       ', dd[4]:0:3);
        end;
        {$I-}
        close(tout);
        {$I+}
    end
    else begin
        out:=CSImg_FileOutput.init(ofn);
        if ft then mi.add('EXTEND', 'T', 'there is a BINTABLE extension');

        { plotting  }
        if fwhm >= 1e-3 then
        begin
            fi.pt:=nil; ri.pt:=nil;
            cc.Info('expsources: Plotting sources to '+ofn);
            SImg_PlotSources(cc, fi, sl,mi, 0.5*fwhm, sc);
            SImg_Convert(cc,fi,ri,true,true);
            if ri.pt<>fi.pt then SImg_Free(cc,fi);
            SImg_WFITS(cc,out,ri,mi);
            SImg_Free(cc,ri);
        end
        else SImg_WFITSHeader(cc,out,mi);

        { fits table }
        if ft then
        begin
            { align bottom left pixel to (1,1) }
            mi.reset;
            mi.add('IMAGEW', int2str(round(xmax-xmin+1)), 'Image width');
            mi.add('IMAGEH', int2str(round(ymax-ymin+1)), 'Image height');
            { transform computer coordinate system to fits coordinates }
            xmin-=1; ymax+=1;
            for y:=0 to sl.ny-1 do
            begin
                sl.ptd[y*sl.nx  ]-=xmin;
                sl.ptd[y*sl.nx+1]:=ymax-sl.ptd[y*sl.nx+1];
            end;
            cc.Info('expsources: Writing source list to '+ofn);
            SImg_AFITSTable(cc,out,sl,['X','Y','Flux','Background','Weight'],mi);
        end;
        out.close(cc);
    end;

    SImg_Free(cc,sl);
    mi.destroy;
    cc.destroy;
end.
