{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name fwhmmap}
{@description Calculates a star FWHM map that can be used for sdbadd}

{$I directives.pas}
uses simage,getparam;
var par           : tparams;
    cc            : CSImg_Config;
    mi            : CSImg_MetaInfo;
    fi            : TSImg_FPImage;
    ifn,ofn,val   : ansistring;
    bps           : shortstring;
    rad           : longint;
    res,dmin      : double;
    bi            : TSImg_BayerInfo;

begin
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        ifn:='';      defparam('<infile>','Source image (default: stdin)',ifn);
        ofn:='';      defparam('<outfile>','Destination image (default: stdout)',ofn);
                      defparam('-h','Help',helpf);
        dmin:=2;      defparam('-m','Minimum FWHM d_min. Default: 2 (As larger as better the noise immunity).\\'+
                                    'Values less then 1.5 should be avoided)',dmin);
        rad:=-1;      defparam('-r','Star search radius. Recommended: at least 1.5 times the FWHM. Default: d_min*10',rad);
        res:=4;       defparam('-s','Weight smoothing relative to image dimension. Recommended: 2..10, Default: 4',res);
        bps:='';      defparam('-b','Bayer pattern (default: auto-detect)',bps);
                      defparam('-o','Destination image',ofn); deffmt('-o','<outfile>');
        run;
        done;
    end;
    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug('fwhmmap: values: source='+ifn+' dest='+ofn+' dmin='+real2str(dmin)+' rad='+int2str(rad)+' smoothing='+real2str(res)+' bp='+bps);

    mi:=CSImg_MetaInfo.create();
    if dmin<0.7 then dmin:=0.5;
    if rad<dmin then rad:=round(dmin*10+0.49);
    cc.Info('fwhmmap: Reading input image '+ifn);
    fi.pt:=nil;
    SImg_ReadImageFile(cc,ifn,fi,mi);

    if (bps='') and mi.get('BAYERPAT',val) then bps:=trimquotes(val);
    if bps<>'' then
    begin
        bi.cf[0]:=1;
        bi.cf[1]:=1;
        bi.cf[2]:=1;
        bi.cf[3]:=1;
        bi.pattern:=SImg_BayerParsePattern(cc,bps);
        if not SImg_BayerCheckInfo(bi) then cc.warning('info: Invalid Bayer pattern')
        else begin
            cc.Info('info: Converting Bayer pattern image');
            SImg_BayerDownscale(cc,bi,fi,fi);
        end;
    end;

    cc.Info('fwhmmap: Calculating FWHM using ' +int2str(cc.threads) + ' threads');
    res:=SImg_FWHMMap(cc, fi,fi, dmin,rad,res);
    cc.info('fwhmmap: Total sharpness: '+real2str(res,-1)+' pixel (FWHM)');

    mi.reset();
    mi.add('FWHMPIX',real2str(res,-4),'Star FWHM in pixels');
    cc.Info('fwhmmap: Writing output image '+ofn);
    SImg_WriteImageFile(cc,ofn,fi,mi);

    SImg_Free(cc,fi);
    mi.destroy;
    cc.destroy;
end.

