{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@ame getalignment}
{@description Calculates alignment of a source list (usually stars from an image)}
{@longdescription
Calculates how a the reference source list must be aligned to fit to destination source list.
With the parameter -d the expected distortion can be specified 0 is recommended for telescopes,
1 is the best choice for standard and telephoto lenses and 2 should be used for wide angle lenses and
3 can be tried for fish-eye lenses. This setting is only relevant for finding the initial
matches.
  Constraints for the initial solution can be specified using -nq, -smin, -smax, -dmax, dx and -dy.
}

{$I directives.pas}
uses simage,getparam,sysutils;
var par                 : tparams;
    cc                  : CSImg_Config;
    fir,fid,fo,msg      : ansistring;
    rmi,dmi             : CSImg_MetaInfo;
    pr,pd               : TSImg_RDImage;
    dl                  : longint;
    res                 : TSImg_AlignmentParams;
    con                 : TSImg_AlignmentConstraints;
    wcs                 : double;
    mo                  : boolean;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='<infile1> [<infile2> [<outfile>]]';
        fir:='';         defparam('<infile1>','Reference source list',@fir,tpastr,rapdreq);
        fid:='';         defparam('<infile2>','Destination source list (default: stdin)',fid);
        fo:='';          defparam('<outfile>','Source list containing alignment information (default: <infile2> if defined, otherwise stdout)',fo);
        dl:=1;           defparam('-d','Distortion level (0=low .. 3=extreme, default: 1)',dl);
        con.nq:=false;   defparam('-nq','Allow non-quadratic pixels', con.nq);
        con.flip:=false; defparam('-f','Allow flipping', con.flip);
        con.smin:=0.1;   defparam('-smin','Minimum scale (default: 0.1)', con.smin);
        con.smax:=10;    defparam('-smax','Maximum scale (default: 10)', con.smax);
        con.dmax:=1e20;  defparam('-dmax','Maximum shift distance of the reference point given by -rx and -ry from origin (center) in destination', con.dmax);
        con.rx:=0;       defparam('-rx','X coordinate in reference which is transformed to origin (center) of destination (default: 0)', con.rx);
        con.ry:=0;       defparam('-ry','Y coordinate in reference which is transformed to origin (center) of destination (default: 0)', con.ry);
        con.sh:=-1;      defparam('-mc','Maximum shearing angle in degree. Default: auto', con.sh);
        con.frc:=0;      defparam('-frc','Field rotation in degree. Default: 0', con.frc);
        con.frd:=200;    defparam('-frd','Maximum difference of field rotation from value given by -rc in degree. Default: 180', con.frd);
        wcs:=-1;         defparam('-w','If WCS data is given, use them to calculate constraints with a given tolerance (recommended: 0.05..0.25)', wcs);
        mo:=false;       defparam('-m','Only write meta info', mo);
                         defparam('-h','Help',helpf);
                         defparam('-o','Source list containing alignment information',fo); deffmt('-o','<outfile>');
        run;
        done;
    end;

    cc.debug('getalignment: in_ref='+fir+' in_dest='+fid+' out='+fo+'  dl='+int2str(dl));

    { reference source list }
    pr.pt:=nil;
    rmi:=CSImg_MetaInfo.create();
    cc.Info('getalignment: Reading reference source list '+fir);
    SImg_ReadImageFile(cc,fir,pr,rmi);

    { destination source list }
    pd.pt:=nil;
    dmi:=CSImg_MetaInfo.create();
    cc.Info('getalignment: Reading destination source list '+fid);
    SImg_ReadImageFile(cc,fid,pd,dmi);
    SImg_ReadAlignmentParams(cc,res,dmi);

    if wcs>=1e-10 then
    begin
        msg:=SImg_GetAlignmentConstraints(cc,rmi,dmi, wcs, con);
        if msg<>'' then cc.warning('Unable to calculate constraints from WCS data: '+msg);
        cc.info('getalignment: reference_pos=('+real2str(con.rx,0)+', '+real2str(con.ry,0)+')'+'  max_shift='+real2str(con.dmax,0)+'  flip='+boolean2str(con.flip)+'  scale='+real2str(con.smin)
        + '...' + real2str(con.smax) + '  nq='+boolean2str(con.nq)+'  shearing='+real2str(con.sh,-1)+'  field_rotation='+real2str(con.frc,-1)+' +/- '+real2str(con.frd,-1) );
    end;

    { sources }
    SImg_GetAlignment(cc, pr,pd, dl, res, con);
    res.weight:=1;

    { write output  }
    if fo='' then fo:=fid;
    cc.Info('getalignment: Writing alignment data '+fo);
    SImg_WriteAlignmentParams(res,dmi);
    if mo then SImg_WFITSHeader(cc,CSImg_FileOutput.init(fo),dmi).close(cc)
    else SImg_WFITS(cc,CSImg_FileOutput.init(fo),pd,dmi).close(cc);

    rmi.destroy;
    dmi.destroy;
    SImg_Free(cc,pr);
    SImg_Free(cc,pd);
    cc.destroy;
end.
