{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name starsubtract}
{@description Extract stars form image and subtract or interpolate them.}

{$I directives.pas}
uses simage,getparam,sysutils;

var par                 : tparams;
    cc                  : CSImg_Config;
    ifn,mfn,ofn         : ansistring;
    omfn,osfn,opfn,odfn : ansistring;
    opafn               : ansistring;
    fi,fs,fb,fd         : TSImg_FPImage;  { tile images }
    fpsf,fpsfs          : TSImg_FPImage;  { tile PSF, PSF sum }
    fim,fsm,fbm,fdm     : TSImg_FPImage;  { mosaic images }
    rmm,rmo,rm          : TSImg_RDImage;
    tol,tolc,ff,t1,t2   : double;
    ithrn,ithrr,iwdt    : double;
    cwdt,sym,bf,sig     : double;
    fwhm,sym2,rrad      : double;
    satthr              : double;
    rad,getmask         : longint;
    ma,mf               : boolean;
    x,y,opfnx,opfny     : longint;
    ti                  : TSImg_TileInfo;
    name                : shortstring;
    mi,mj               : CSImg_MetaInfo;
    shm,sh              : TSImg_StarSubtract_shared;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);

    with par do
    begin
        init;
        paramusage:='[<infile>] [<outfile>]';
        ifn:='';     defparam('<infile>', 'Source image (default: stdin)', ifn);
        ofn:='';     defparam('<outfile>','Destination image (default: stdout)', ofn);
        mfn:='';     defparam('-m','Input star mask (bit 1 and 2 / mask 2 and 4) as written by -om.', mfn);  deffmt('-m','<infile>');
        ma:=false;   defparam('-ma','Allow to find new stars if -m is given.',ma);
        mf:=false;   defparam('-mf','Filter stars given by -m. By default they are treated as save stars.',mf);
        rad:=100;    defparam('-r','PSF radius: Default: 100', rad);
        rrad:=1;     defparam('-rr','Use global PSF above that relative radius. Default: 1. Ignored if only one tile', rrad);
        sym:=1;      defparam('-s','Symmetry enforcement for global PSF. Valid: 0 (non-symmetric none) .. 1 (fully symmetric). Default: 1', sym);
        sym2:=0;     defparam('-st','PSF symmetry enforcement for tiles. Ignored if only one tile. Default: 0', sym2);
        tol:=1.4;    defparam('-t','PSF tolerance as factor in center. Default: 1.4', tol);
        tolc:=0;     defparam('-tc','PSF tolerance as factor in corners. Default: same as -t. Only used for global PSF calculation if more than on tiles.', tolc);
        ithrn:=10;   defparam('-i','Interpolation threshold relative to noise level. Default: 10', ithrn);
        ithrr:=1;    defparam('-j','Interpolation threshold relative to residual. Default: 1', ithrr);
        iwdt:=4;     defparam('-w','Interpolation width relative to FWHM. Valid: >1. Recommended: >2. Default: 4', iwdt);
        cwdt:=0.75;  defparam('-c','Maximum PSF correction width relative to PSF size. Valid: 0..1, Default 0.75. Recommended: 0.5..1', cwdt);
        ff:=0.5;     defparam('-f','Filter strength. (Default: 0.5. Smaller values to consider more objects as stars.', ff);
        bf:=1;       defparam('-b','Noise correction factor. Default and recommended: 1. Detection threshold can be increased manually if noise is not constant.', bf);
        t1:=1;       defparam('-n1','Split into this amount of tiles into direction of larger dimension for overall calculation. (default: 1)',t1);
        t2:=3;       defparam('-n2','Split into this amount of tiles into direction of larger dimension for PSF correction. <=0 to disable. Default: 3',t2);
        fwhm:=-1;    defparam('-fwhm','Input star FWHM. 0 for auto-detect. Default: try to read from image, otherwise auto-detect.', fwhm);
        satthr:=0.5; defparam('-u','Saturation threshold relative to brightest pixel. Default: 0.5. Reduce this value if bright stars are not detected.', satthr);
                     defparam('-o','Output image',ofn);                         deffmt('-o','<outfile>');
        omfn:='';    defparam('-om','Output mask: bit0: interpolation, bit 1: star mask, bit 2: save star',omfn);       deffmt('-om','<outfile>');
        osfn:='';    defparam('-os','Output deconvoluted star image',osfn);                                             deffmt('-os','<outfile>');
        odfn:='';    defparam('-od','Output difference, i.e. convoluted and extrapolated stars',odfn);          deffmt('-od','<outfile>');
        opfn:='';    defparam('-op','Output PSF. %x and %y are replaced by tile indices.', opfn);                       deffmt('-op','<outfile>');
        opafn:='';   defparam('-opa','Output global PSF (makes only sense if n1>1)', opafn);                          deffmt('-opa','<outfile>');
                     defparam('-h','Help',helpf);
        run;
        done;
    end;
    cc.debug('starsubtract: source='+ifn+' mask_in='+mfn + ' out='+ofn+ ' star_out='+osfn+ ' mask_out='+omfn+' psf_out='+' psfa_out='+opafn+' diff_out='+odfn+
            ' tol='+real2str(tol)+' tolc='+real2str(tolc)+' rad='+real2str(rad)+' ithrn='+real2str(ithrn)+' ithrr='+real2str(ithrr)+' iwdt'+real2str(iwdt,-2)+' cwdt'+real2str(cwdt,-2)+
            ' ff='+real2str(ff)+' bf='+real2str(bf)+' n1='+realstr(t1,-1)+' n2='+real2str(t2,-1)+' ma='+boolean2str(ma)+' sym='+real2str(sym)+' sym2='+real2str(sym2)+' fwhm='+real2str(fwhm)+
            ' rrad='+real2str(rrad)+' satthr='+real2str(satthr));
    if (ofn='') and (omfn='') and (osfn='') and (opfn='') and (opafn='') then
    begin
        cc.warning('No output defined.');
        SImg_Halt;
    end;

    if rad<10 then rad:=10;
    name:=cc.name;
    if t1<1 then t1:=1;
    if tolc<tol then tolc:=tol;

    { input image }
    fim.pt:=nil;
    mi:=CSImg_MetaInfo.create();
    cc.Info(cc.name+': Reading input image '+ifn);
    SImg_ReadImageFile(cc,ifn,fim, mi);
    mi.deleteRaw;
    if fim.nc<>1 then cc.error(cc.name+': Only one color images are supported.');
    if (fwhm<-0.499) and mi.get('FWHMPIX',fwhm) then cc.info(cc.name+': Read star FWHM from input image: '+real2str(fwhm,-2))
    else fwhm:=0;
    if fwhm<0.499 then
    begin
        fwhm:=SImg_GetSharpnessFWHM(cc,fim);
        cc.Info(cc.name+': Autodetected star FWHM: '+real2str(fwhm,-2));
    end;

    { shared data }
    cc.info(cc.name+': Background subtraction and calculation of weight image.');
    SImg_SubtractBackground(cc, fim,shm, rad,fwhm,ithrn,iwdt,satthr);
    if bf<0.1 then bf:=0.1;
    shm.sig*=bf;

    { tiles }
    if fim.nx>fim.ny then
    begin
        ti.nx:=round(t1);
        ti.ny:=round(fim.ny/fim.nx*t1);
    end
    else begin
        ti.ny:=round(t1);
        ti.nx:=round(fim.nx/fim.ny*t1);
    end;
    if ti.ny<1 then ti.ny:=1;
    if ti.nx<1 then ti.nx:=1;
    ti.bb:=rad;
    ti.bu:=rad;
    if (ti.nx>1) or (ti.ny>1) then cc.info(name+': Splitting image into '+int2str(ti.nx)+' x '+int2str(ti.ny)+' tiles');
    opfnx:=0;
    opfny:=0;

    if opfn<>'' then
    begin
        opfnx:=pos('%x',opfn); if opfnx<0 then opfnx:=0;
        opfny:=pos('%y',opfn); if opfny<0 then opfny:=0;
        if ((ti.nx>1) and (opfnx<=0)) or ((ti.ny>1) and (opfny<=0)) then cc.error(name+': Multiple PSF''s cannot be written to single file. Use %x and %y.');
        if opfnx>opfny then
        begin
            if opfnx>0 then delete(opfn,opfnx,2);
            if opfny>0 then
            begin
                delete(opfn,opfny,2);
                opfnx-=2;
            end;
        end
        else begin
            if opfny>0 then delete(opfn,opfny,2);
            if opfnx>0 then
            begin
                delete(opfn,opfnx,2);
                opfny-=2;
            end;
        end;
    end;

    { get global PSF }
    fpsfs.pt:=nil; rmm.pt:=nil;
    SImg_GetPSF(cc, shm, fpsfs, tolc, rad, rmm,0, sym,-1);
    fwhm:=SImg_GetPSFWidth(cc,fpsfs, 0.25)*sqrt(0.5);

    {output global PSF }
    if opafn<>'' then
    begin
        mj:=CSImg_MetaInfo.create();
        mj.add('FWHMPIX',real2str(fwhm,-4), 'Star FWHM in pixels');
        cc.Info(cc.name+': Writing global PSF '+opafn);
        SImg_WriteImageFile(cc,opafn,fpsfs,mj);
        mj.destroy;
    end;

    { input mask }
    getmask:=1;
    rmm.pt:=nil;
    if mfn<>'' then
    begin
        cc.Info(cc.name+': Reading mask '+mfn);
        SImg_ReadImageFile(cc,mfn,rmm);
        if (rmm.fmt<>SImg_RawByte) and (rmm.fmt<>SImg_RawFP) then cc.error(cc.name+': Invalid format of star mask');
        if (rmm.nx<>fim.nx) and (rmm.ny<>fim.ny) or (rmm.nc<>1) then cc.error(cc.name+': Invalid size of star mask');
        if not ma then getmask:=0;
        if mf then getmask:=getmask or 2;
    end;

    { process tiles }
    sh:=shm; sh.fi2.pt:=nil; sh.fw.pt:=nil;
    fi.pt:=nil; rm.pt:=nil; fpsf.pt:=nil;       { tiles images }
    fbm.pt:=nil; fsm.pt:=nil; fdm.pt:=nil;      { output mosaic images }
    fpsf.pt:=nil;
    rmo:=rmm;                                   { output mask }
    sig:=0;
    if (ti.nx=1) and (ti.ny=1) then tolc:=tol;
{if (ti.nx>1) or (ti.ny>1) then
begin
    x:=2; y:=1;
//    x:=2; y:=1;
end
else begin
    x:=0; y:=0;
end;  }
    for y:=0 to ti.ny-1 do
        for x:=0 to ti.nx-1 do
        begin
            if (ti.nx>1) or (ti.ny>1) then cc.name:=name+'('+int2str(x+1)+','+int2str(y+1)+')';
            SImg_GetTile(cc, ti, fim,    fi,     x,y);
            SImg_GetTile(cc, ti, shm.fi2,sh.fi2, x,y);
            SImg_GetTile(cc, ti, shm.fw, sh.fw,  x,y);
            if rmm.pt<>nil then SImg_GetTile(cc, ti, rmm,rm, x,y)
            else rm.pt:=nil;

            { calculate PSF }
            if (ti.ny=1) and (ti.nx=1) then fpsf:=fpsfs
            else begin
                SImg_Duplicate(cc, fpsfs, fpsf,true);
                SImg_GetPSF(cc, sh, fpsf, tol, rad, rm,0, sym2,rrad);
            end;

            { output psf }
            if opfn<>'' then
            begin
                ifn:=opfn;
                if opfnx>opfny then insert(int2str(x+1),ifn,opfnx);
                if opfny>0 then insert(int2str(y+1),ifn,opfny);
                if (opfnx>0) and (opfnx<opfny) then insert(int2str(x+1),ifn,opfnx);
                mj:=CSImg_MetaInfo.create();
                mj.add('FWHMPIX',real2str(Simg_GetPSFWidth(cc,fpsf,0.25)*sqrt(0.5)), 'Star FWHM in pixels');
                cc.Info(cc.name+': Writing psf '+ifn);
                SImg_WriteImageFile(cc,ifn,fpsf,mj);
                mj.destroy;
            end;

            if (omfn<>'') or (ofn<>'') or (osfn<>'') or (odfn<>'') then
            begin
                { find stars }
                fb.pt:=nil; fs.pt:=nil; fd.pt:=nil;
{ Subtract stars
    fi     : input (1 color image)
    fpsf   : input PSF
    rms    : mask input and output. Possible inputs:
        1. undefined
        2. mask
            bit 0: interpolated pixels (output only)
            bit 1: star (input and output)
            bit 2: save star (input and output )
        3. floating point (SImg_RawFP) containing deconvoluted stars that will be brightness filtered
    fob    : output image: background
    fos    : output image: deconvoluted stars
    fod    : output image: difference between input and fob where saturated regions are extrapolated
    ofob   : true: output fob; false: do not touch fob
    ofos   : true: output fos; false: do not touch fos
    ofod   : true: output fod; false: do not touch fod
    getmask: bit 0: find new stars
             bit 1: filter input candidates, otherwise they will be set to save stars.
    tol    : psf width tolerance as factor in center
    tolc   : psf width tolerance as factor in corners
    ff     : Source filter strength >=0, high values ==> less objects are considered as stars
    nf     : Noise correction factor. Usually 1.
    ithrn  : Interpolation threshold relative to noise >=1, recommended: >4
    ithrr  : Interpolation threshold relative to residual
    cwdt   : Maximum PSF correction width relative to PSF. Valid: 0..1. Recommended: 0.25..1
    tn     : PSF correction tiles into direction of larger dimension; <0.5 to disable PSF correction
}
                SImg_StarSubtract(cc,  fi, fpsf, sh, rm,  fb, fs, fd, ofn<>'', osfn<>'', odfn<>'',  getmask,  tol, tolc,  ff,  ithrn,ithrr, cwdt,  t2);
                sh.fi2.pt:=nil; sh.fw.pt:=nil;  { because they are freed by SImg_StarSubtract which gets a copy of sh }
                SImg_Free(cc,fi);
                sig+=sqr(sh.sig);

                if (ti.nx>1) or (ti.ny>1) then
                begin
{SImg_WriteImageFile(cc,'starsubtract-fpsf.fits',fpsf);
SImg_WriteImageFile(cc,'starsubtract-rm.fits',rm);
if ofn <>'' then SImg_WriteImageFile(cc,'starsubtract-fb.fits',fb);
if osfn<>'' then SImg_WriteImageFile(cc,'starsubtract-fs.fits',fs);
if odfn<>'' then SImg_WriteImageFile(cc,'starsubtract-fd.fits',fd);
cc.error('no error');  }
                    if fb.pt<>nil then
                    begin
                        if fbm.pt=nil then SImg_New(cc, fbm, fim.nx,fim.ny,1);
{                       for i:=0 to fb.size-1 do
                            fb.pt[i]*=1+(x+y)*0.5; }
                        SImg_BlendTile(cc, ti, fb,fbm, x,y);
                        SImg_Free(cc,fb);
                    end;
                    if fs.pt<>nil then
                    begin
                        if fsm.pt=nil then SImg_New(cc, fsm, fim.nx,fim.ny,1);
                        SImg_CopyTile(cc, ti, fs,fsm, x,y);
                        SImg_Free(cc,fs);
                    end;
                    if fd.pt<>nil then
                    begin
                        if fdm.pt=nil then SImg_New(cc, fdm, fim.nx,fim.ny,1);
                        SImg_BlendTile(cc, ti, fd,fdm, x,y);
                        SImg_Free(cc,fd);
                    end;
                    if rmo.pt=rmm.pt then rmo.pt:=nil;
                    if rmo.pt=nil then SImg_New(cc, rmo, SImg_RawByte, fim.nx,fim.ny,1, true);
                    SImg_CopyTile (cc, ti, rm,rmo, x,y);
                    SImg_Free(cc,rm);
                end
                else begin
                    SImg_Move(cc, fb,fbm);
                    SImg_Move(cc, fs,fsm);
                    SImg_Move(cc, fd,fdm);
                    SImg_Move(cc, rm,rmo);
                end;
            end
            else begin
                SImg_Free(cc,rm);
                SImg_Free(cc,fi);
            end;
            if fpsf.pt<>fpsfs.pt then SImg_Free(cc,fpsf);
        end;
    if (ti.nx>1) or (ti.ny>1) then
    begin
        SImg_Free(cc, fim);
        SImg_Free(cc, shm.fi2);
        SImg_Free(cc, shm.fw);
    end;
    cc.name:=name;
    SImg_Move(cc, rmo,rmm);
    SImg_Free(cc, fpsfs);

    { add meta info }
    sig:=sqrt(sig/ti.nx*ti.ny);
    mi.add('NOISEFLR', real2str(sig,10),'Noise floor');
    mi.add('FWHMPIX',real2str(fwhm,-4), 'Star FWHM in pixels');
    cc.info(cc.name+': Added meta info: NOISEFLR='+real2str(sig)+'  FWHMPIX='+real2str(fwhm,-2));

    { output image }
    if ofn<>'' then
    begin
        cc.Info(cc.name+': Writing output image '+ofn);
        SImg_WriteImageFile(cc,ofn,fbm,mi);
    end;
    SImg_Free(cc, fbm);

    { output deconvoluted stars }
    if osfn<>'' then
    begin
        cc.Info(cc.name+': Writing deconvoluted star image '+osfn);
        SImg_WriteImageFile(cc,osfn,fsm,mi);
    end;
    SImg_Free(cc, fsm);

    { output convoluted stars }
    if odfn<>'' then
    begin
        cc.Info(cc.name+': Writing difference / convoluted star image '+osfn);
        SImg_WriteImageFile(cc,odfn,fdm,mi);
    end;
    SImg_Free(cc, fdm);

    { output mask }
    if omfn<>'' then
    begin
        cc.Info(cc.name+': Writing mask '+omfn);
        SImg_WriteImageFile(cc,omfn,rmm,mi);
    end;
    SImg_Free(cc, rmm);

    cc.destroy;
end.
