{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******** SImg_Noise *****************************************************}
{*************************************************************************}
{ Generate rectangular-distributed noise on interval [-0.5,0.5]
   nx,ny,nc  : dimensions }
procedure SImg_Noise_Rect(cc:CSImg_Config; var fi: TSImg_FPImage; nx,ny,nc:longint);
var i  : longint;
    d  : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Noise_rect: fi='+SImg_InfoStr2(fi)+' nx='+int2str(nx)+' ny='+int2str(ny)+' nc='+int2str(nc));
    SImg_New(cc,fi,nx,ny,nc);
    d:=fi.pt;
    for i:=0 to fi.size-1 do
    begin
        d^:=random-0.5;
        d+=1;
    end;
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******** SImg_Noise *****************************************************}
{*************************************************************************}
{ Generate normal-distributed noise (sigma=1, mu=0)
   nx,ny,nc  : dimensions }
procedure SImg_Noise_Norm(cc:CSImg_Config; var fi: TSImg_FPImage; nx,ny,nc:longint);
const tres = 65536;
      dt   = 1e-6;
var trans         : array[0..tres-1] of single;
    i,j           : longint;
    r,s,t,u,v1,v2 : double;
    d             : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Noise_norm: fi='+SImg_InfoStr2(fi)+' nx='+int2str(nx)+' ny='+int2str(ny)+' nc='+int2str(nc));

    u:=1/sqrt(2*pi);
    s:=0; t:=0; v1:=u;
    for i:=0 to tres-1 do
    begin
        r:=(i+0.5)/tres;
        while s<r do
        begin
            t+=dt;
            v2:=exp(-0.5*sqr(t))*u;
            s+=(v1+v2)*dt;
            v1:=v2;
        end;
        trans[i]:=t-0.5*dt;
    end;

    SImg_New(cc,fi,nx,ny,nc);
    d:=fi.pt;
    for i:=0 to fi.size-1 do
    begin
        j:=random(2*tres+1);
        if j<tres then d^:=-trans[tres-1-j]
        else if j>tres then d^:=trans[j-tres-1]
        else d^:=0;
        d+=1;
    end;

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******** SImg_Gauss *****************************************************}
{*************************************************************************}
{ Draw 3-dimensional Gauss curve (2-dimensional if nc=1)
   nx,ny,cs  : dimensions
   sx,sy,sc  : sigma
   re        : radius exponent (1 for Gauss curve) }
procedure SImg_Gauss(cc:CSImg_Config; var fi: TSImg_FPImage; nx,ny,nc:longint; sx,sy,sc, re:double);
var x,y,c        : longint;
    xc,yc,zc,r,s : double;
    d            : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Gauss: fi='+SImg_InfoStr2(fi)+' nx='+int2str(nx)+' ny='+int2str(ny)+' nc='+int2str(nc)+'   sx='+real2str(sx)+' sy='+real2str(sy)+' sc='+real2str(sc));
    if abs(sx)<1e-50 then
    begin
        if sx<0 then sx:=-1e-50
        else sx:=1e-50;
    end;
    if abs(sy)<1e-50 then
    begin
        if sy<0 then sy:=-1e-50
        else sy:=1e-50;
    end;
    if abs(sc)<1e-50 then
    begin
        if sc<0 then sc:=-1e-50
        else sc:=1e-50;
    end;
    sx:=1/sqr(sx);
    sy:=1/sqr(sy);
    sc:=1/sqr(sc);
    xc:=(nx-1)*0.5;
    yc:=(ny-1)*0.5;
    zc:=(nc-1)*0.5;

    SImg_New(cc,fi,nx,ny,nc);

    d:=fi.pt;
    s:=1e-100;
    for y:=0 to ny-1 do
        for x:=0 to nx-1 do
            for c:=0 to nc-1 do
            begin
                r:=(sqr(x-xc)*sx + sqr(y-yc)*sy + sqr(c-zc)*sc);
                if r>1e-100 then r:=exp(-0.5*exp(ln(r)*re))
                else r:=1;
                s+=r;
                d^:=r;
                d+=1;
            end;

    s:=1/s;
    d:=fi.pt;
    for x:=0 to fi.size-1 do
    begin
        d^*=s;
        d+=1;
    end;

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******** SImg_MexHat ****************************************************}
{*************************************************************************}
{ Draw 3-dimensional Mexican hat curve (2-dimensional if nc=1)
   nx,ny,cs  : dimensions
   sx,sy,sc  : sigma }
procedure SImg_MexHat(cc:CSImg_Config; var fi:TSImg_FPImage; nx,ny,nc:longint; sx,sy,sc:double);
var i,j,k    : longint;
    r        : double;
    xc,yc,zc : double;
    x,y,c    : double;
    d        : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_Gauss: fi='+SImg_InfoStr2(fi)+' nx='+int2str(nx)+' ny='+int2str(ny)+' nc='+int2str(nc)+'   sx='+real2str(sx)+' sy='+real2str(sy)+' sc='+real2str(sc));
    if abs(sx)<1e-50 then
    begin
        if sx<0 then sx:=-1e-50
        else sx:=1e-50;
    end;
    if abs(sy)<1e-50 then
    begin
        if sy<0 then sy:=-1e-50
        else sy:=1e-50;
    end;
    if abs(sc)<1e-50 then
    begin
        if sc<0 then sc:=-1e-50
        else sc:=1e-50;
    end;
    sx:=1/sqr(sx);
    sy:=1/sqr(sy);
    sc:=1/sqr(sc);
    xc:=(nx-1)*0.5;
    yc:=(ny-1)*0.5;
    zc:=(nc-1)*0.5;

    SImg_New(cc,fi,nx,ny,nc);
    d:=fi.pt;
    for i:=0 to ny-1 do
        for j:=0 to nx-1 do
            for k:=0 to nc-1 do
            begin
                x:=sqr(j-xc)*sx;
                y:=sqr(i-yc)*sy;
                c:=sqr(k-zc)*sc;
                r:=(1-x-y-c)*exp(-0.5*(x+y+c));
                d^:=r;
                d+=1;
            end;
end;

