{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******* SImg_AvgSet *****************************************************}
{*************************************************************************}
{
   Calculate avarage and noise of an image set
   iset: image set
   fi  : avg output
   fn  : square noise output
   d   : drop brightest and darkest value
   dmin: minimum valid value
   dmax: maximum valid value
}
procedure SImg_AvgSet(cc:CSImg_Config; iset:CSImg_ImageSet; var ofi,ofn:TSImg_FPImage; d:boolean; dmin,dmax:double);
var fi,fj,fk,fl,fm  : TSImg_FPImage;
    wi              : ^word;
    i,ia,j          : longint;
    r,s,min,max     : simgfp;
    u               : double;
    w               : word;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_AvgSet: called with: iset='+iset.info+' d='+boolean2str(d));
    iset.cnx:=true;
    iset.cny:=true;
    iset.cnc:=true;
    if iset.getmax<0 then cc.error('SImg_AvgSet: No images to add');
    if d and (iset.getmax<2) then cc.error('SImg_AvgSet: At least 3 images required');

    SImg_Free(cc,ofi);
    SImg_Free(cc,ofn);

    fi.pt:=nil; fj.pt:=nil;
    fk.pt:=nil; fl.pt:=nil; fm.pt:=nil;
    ia:=0;

    // init min max buffers
    min:=simg_large;
    max:=-simg_large;
    if d then
    begin
        ia:=2;
        iset.getimage(cc,fl,0);
        iset.getimage(cc,fm,1);
        for j:=0 to fl.size-1 do
        begin
            r:=fl.pt[j];
            s:=fm.pt[j];
            if r>s then
            begin
                fm.pt[j]:=r;
                fl.pt[j]:=s;
                if r>max then max:=r;
                if s<min then min:=s;
            end else
            begin
                if s>max then max:=s;
                if r<min then min:=r;
            end;
        end;
    end;

    // add images
    for i:=ia to iset.getmax do
    begin
        iset.getimage(cc,fk,i);
        if i=ia then
        begin
            SImg_Duplicate(cc,fk,fi,false);
            SImg_Duplicate(cc,fk,fj,false);
            getmem2(wi,fk.size*2);
            fillchar(wi^,fk.size*2,0);
        end;
        for j:=0 to fk.size-1 do
        begin
            r:=fk.pt[j];
            if d then
            begin
                s:=r;
                if s<fl.pt[j] then
                begin
                    r:=fl.pt[j];
                    fl.pt[j]:=s;
                end;
                if s>fm.pt[j] then
                begin
                    r:=fm.pt[j];
                    fm.pt[j]:=s;
                end;
            end;
            if (r>=dmin) and (r<=dmax) then
            begin
                inc(wi[j]);
                fi.pt[j]+=r;
                fj.pt[j]+=sqr(r);
            end
            else if wi[j]=0 then fi.pt[j]:=r;
            if r>max then max:=r;
            if r<min then min:=r;
        end;
    end;

    SImg_Free(cc,fk);
    if d then
    begin
        SImg_Free(cc,fl);
        SImg_Free(cc,fm);
    end;

    s:=sqr(0.5*(max-min));
    for j:=0 to fi.size-1 do
    begin
        w:=wi[j];
        if w>=1e-3 then
        begin
            u:=fi.pt[j]/w;
            fi.pt[j]:=u;                // avg
            if w>=1.0001 then
            begin
                u:=(fj.pt[j]-w*sqr(u))/(w-1);
                if u<simg_mina then u:=simg_mina;
                if u<simg_mina then u:=simg_mina;
            end
            else u:=simg_large;
            fj.pt[j]:=u;                // square noise
        end
        else fj.pt[j]:=s;
    end;

    freemem2(wi,fi.size*2);
    ofi:=fi;
    ofn:=fj;

    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_GetNoise ***************************************************}
{*************************************************************************}
{ get (read out) noise from an (bias) image set }
function SImg_GetNoise(cc:CSImg_Config; iset:CSImg_ImageSet): double;
var fi,fj,fk     : TSImg_FPImage;
    i,j,k        : longint;
    cn,cp        : int64;
    r,s,t        : double;

begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_GetNoise: called with: iset='+iset.info);
    iset.cnx:=true;
    iset.cny:=true;
    iset.cnc:=true;
    if iset.getmax<1 then cc.error('SImg_GetNoise: At least 2 images required');

    fk.pt:=nil; fi.pt:=nil; fj.pt:=nil;
    iset.getimage(cc,fk,0);
    SImg_Duplicate(cc,fk,fi,false);
    SImg_Duplicate(cc,fk,fj,false);

    { step 1: 1st and 2nd moment, average }
    cn:=0; cp:=0; s:=0;
    for i:=0 to iset.getmax do
    begin
        if i>0 then iset.getimage(cc,fk,i);
        for j:=0 to fk.size-1 do
        begin
            r:=fk.pt[j];
            if r<-0.1 then cn+=1
            else if r>0.1 then cp+=1;
            s+=r;
            fi.pt[j]+=r;
            fj.pt[j]+=sqr(r);
        end;
    end;
    k:=iset.getmax+1;
    if (cn<0.25*cp) and (cp<(iset.getmax+1)*(fi.size*0.6)) then
    begin
        cc.warning('SImg_GetRDNoise: Negative values seem to be cropped, result will only be accurate if image is unbiased');
        s:=sqr(s/fi.size);
        k*=2;
    end
    else s:=0;

    { calculate noise }
    r:=0;
    for j:=0 to fi.size-1 do
    begin
        t:=fj.pt[j]*k - sqr(fi.pt[j]) + s;
        if t>0 then r+=t;
    end;
    result:=sqrt(r/(fi.size*sqr(iset.getmax+1.0)) * (1+1/(iset.getmax+1)));

    SImg_Free(cc,fi);
    SImg_Free(cc,fj);
    SImg_Free(cc,fk);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_GetGain ****************************************************}
{*************************************************************************}
{ get gain from an image set
    fb  : bias image (optional, will be cropped to ROI)
    mi  : meta info. Must contain read out noise (unless defeined in input images) and used to crop region of interest
}
function SImg_GetGain(cc:CSImg_Config; iset:CSImg_ImageSet; var fb:TSImg_FPImage; mi:CSImg_MetaInfo): double;
var fi,fj,fk            : TSImg_FPImage;
    i,j                 : longint;
    rdnoise,a,r,s,t     : double;
    r1,r2,s1,s2         : double;

begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_getGaint: called with: iset='+iset.info + ' fb='+SImg_Infostr2(fb));
    if (fb.pt<>nil) and (not SImg_Recover(cc,fb)) then cc.error('SImg_GetGain: Invalid bias image');

    iset.cnx:=true;
    iset.cny:=true;
    iset.cnc:=true;
    if iset.getmax<1 then cc.error('SImg_GetNoise: At least 2 images required');

    fk.pt:=nil; fi.pt:=nil; fj.pt:=nil;
    iset.getimage(cc,fk,mi,0);
    if not mi.get('RDNOISE',rdnoise) then cc.error(cc.name+': Read out noise not defined.');
    if fb.pt<>nil then
    begin
        if (fb.nx<>fk.nx) or (fb.ny<>fk.ny) or (fb.nc<>fk.nc) then cc.error('SImg_GetGain: Dimensions of input image and bias images does not match');
        SImg_CropROI(cc, fb, mi);
    end;
    SImg_CropROI(cc, fk, mi);
    SImg_Duplicate(cc,fk,fi,false);
    SImg_Duplicate(cc,fk,fj,false);

    { step 1: 1st and 2nd moments }
    s1:=0; s2:=0;
    for i:=0 to iset.getmax do
    begin
        if i>0 then
        begin
            iset.getimage(cc,fk,i);
            SImg_CropROI(cc, fk, mi);
        end;

        { subtract bias }
        if fb.pt<>nil then
        begin
            for j:=0 to fk.size-1 do
                fk.pt[j]-=fb.pt[j];
        end;

        { compensate variation in exposure time / focal ratio / brightness / bias }
        s:=0;
        for j:=0 to fk.size-1 do
            s+=fk.pt[j];
        s:=fk.size/s;
        s1+=s;
        s2+=sqr(s);

        for j:=0 to fk.size-1 do
        begin
            r:=fk.pt[j]*s;
            fi.pt[j]+=r;
            fj.pt[j]+=sqr(r);
        end;
    end;

    { step 2: calculate gain }
    s:=sqr(rdnoise)*s2*(iset.getmax+1);
    t:=1+1/(iset.getmax+1);
    r1:=0; r2:=0;
    for j:=0 to fk.size-1 do
    begin
        r:=fi.pt[j];
        a:=((iset.getmax+1)*fj.pt[j] - sqr(r))*t - s;
        if a<0 then a:=0;
        r1+=r*a;
        r2+=sqr(r);
        end;
    result:=r2/r1*s1;
    SImg_Free(cc,fi);
    SImg_Free(cc,fj);
    SImg_Free(cc,fk);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_MinSet *****************************************************}
{*************************************************************************}
threadprocedure SImg_MinSet_thread(fi:TSImg_FPImage; buf:psimgfp; m:longint; n:longint; ia:longint; iz:longint);
var d       : ^simgfp;
    r       : simgfp;
    i,j,ja  : longint;
begin
    d:=fi.pt+ia;
    buf+=n*ia;
    if m<n then ja:=m
    else ja:=n-1;
    for i:=ia to iz do
    begin
        j:=ja;
        r:=d^; d+=1;
        if (m<n) or (r<buf[ja]) then buf[ja]:=r;
        while (j>0) and (r<buf[j-1]) do
        begin
            buf[j]:=buf[j-1];
            buf[j-1]:=r;
            j-=1;
        end;
        buf+=n;
    end;
end;

{ Find the the darkest n=1..8 values of each pixels an return their average }
procedure SImg_MinSet(cc:CSImg_Config; iset:CSImg_ImageSet; var fj:TSImg_FPImage; n:longint);
var fi       : TSImg_FPImage;
    buf      : ^simgfp;
    m,j,tn   : longint;
    d,e      : ^simgfp;
    r,s      : double;
    threads  : PSImg_MinSet_thread;

begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_MinSet: called with: iset='+iset.info + ' out='+SImg_Infostr2(fj)+' n='+int2str(n));
    if (n<1) or (n>8) then  cc.error('SImg_MinSet: Invalid value of n: '+int2str(n));
    iset.cnx:=true;
    iset.cny:=true;
    iset.cnc:=true;
    if iset.getmax<n-1 then cc.error('SImg_GetNoise: At least '+int2str(n)+' images required');

    fi.pt:=nil;
    iset.getimage(cc,fi,0);
    getmem2(buf,fi.size*sizeof(simgfp)*n);

    tn:=cc.threads;
    getmem(threads,tn*sizeof(CSImg_MinSet_thread));
    for j:=0 to tn-1 do
        threads[j]:=CSImg_MinSet_thread.create;

    for m:=0 to iset.getmax do
    begin
        if m>0 then iset.getimage(cc,fi,m);
        for j:=0 to tn-1 do
            threads[j].setwork(fi,buf,m,n, j*fi.size div tn, ((j+1)*fi.size div tn)-1);
        for j:=0 to tn-1 do
            threads[j].wait;
    end;

    for j:=0 to tn-1 do
        threads[j].destroy;
    freemem(threads);

    d:=fi.pt;
    e:=buf;
    s:=1/n;
    for j:=0 to fi.size-1 do
    begin
        r:=0;
        for m:=0 to n-1 do
            r+=e[m];
        d^:=r*s;
        d+=1; e+=n;
    end;

    freemem2(buf,fi.size*sizeof(simgfp)*n);
    SImg_Move(cc, fi,fj);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_MaxSet *****************************************************}
{*************************************************************************}
threadprocedure SImg_MaxSet_thread(fi:TSImg_FPImage; buf:psimgfp; m:longint; n:longint; ia:longint; iz:longint);
var d       : ^simgfp;
    r       : simgfp;
    i,j,ja  : longint;
begin
    d:=fi.pt+ia;
    buf+=n*ia;
    if m<n then ja:=m
    else ja:=n-1;

    for i:=ia to iz do
    begin
        j:=ja;
        r:=d^; d+=1;
        if (m<n) or (r>buf[ja]) then buf[ja]:=r;
        while (j>0) and (r>buf[j-1]) do
        begin
            buf[j]:=buf[j-1];
            buf[j-1]:=r;
            j-=1;
        end;
        buf+=n;
    end;
end;

{ Find the the darkest n=1..8 values of each pixels an return their average
  db : drop brightest value }
procedure SImg_MaxSet(cc:CSImg_Config; iset:CSImg_ImageSet; var fj:TSImg_FPImage; n:longint; db:boolean);
var fi       : TSImg_FPImage;
    buf      : ^simgfp;
    m,j,tn,k : longint;
    d,e      : ^simgfp;
    r,s      : double;
    threads  : PSImg_MaxSet_thread;

begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_MaxSet: called with: iset='+iset.info + ' out='+SImg_Infostr2(fj)+' n='+int2str(n));
    if (n<1) or (db and (n<2)) or (n>8) then  cc.error('SImg_MaxSet: Invalid value of n: '+int2str(n));
    iset.cnx:=true;
    iset.cny:=true;
    iset.cnc:=true;
    if iset.getmax<n-1 then cc.error('SImg_GetNoise: At least '+int2str(n)+' images required');

    fi.pt:=nil;
    iset.getimage(cc,fi,0);
    getmem2(buf,fi.size*sizeof(simgfp)*n);

    tn:=cc.threads;
    getmem(threads,tn*sizeof(CSImg_MaxSet_thread));
    for j:=0 to tn-1 do
        threads[j]:=CSImg_MaxSet_thread.create;

    for m:=0 to iset.getmax do
    begin
        if m>0 then iset.getimage(cc,fi,m);
        for j:=0 to tn-1 do
            threads[j].setwork(fi,buf,m,n, j*fi.size div tn, ((j+1)*fi.size div tn)-1);
        for j:=0 to tn-1 do
            threads[j].wait;
    end;

    for j:=0 to tn-1 do
        threads[j].destroy;
    freemem(threads);

    d:=fi.pt;
    e:=buf;
    if db then k:=1
    else k:=0;

    s:=1/(n-k);
    for j:=0 to fi.size-1 do
    begin
        r:=0;
        for m:=k to n-1 do
            r+=e[m];
        d^:=r*s;
        d+=1; e+=n;
    end;

    freemem2(buf,fi.size*sizeof(simgfp)*n);
    SImg_Move(cc, fi,fj);
    SImg_DecDebugLevel;
end;
