{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

// define[debug_stars]

{ Methods for star FWHM calculation }

{******* SImg_SharpnessFilter_Threshold **********************************}
{ threshold for sharpens filter
  theshold is: <avg> + (<ref_quant-quantile>-<avg>)*<threshold factor>
  fi   : input image
  buf  : threshold and brightest pixel per color
  tf   : threshold factor. Recommended: 0..10 (2..10 for stars)
}
procedure SImg_SharpnessFilter_threshold(cc:CSImg_Config; fi:TSImg_FPImage; buf:pdouble; tf:double);
const ref_quant = 0.98; { reference quantile }
      sat_fact  = 0.9;   // pixels brighter than <brightest pixel>sat_fact are considered as saturated
var d    : ^simgfp;
    r    : simgfp;
    s    : double;
    i,c  : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetSharpnessThreshold2: Invalid image: fi');
    // calculate average
    for c:=0 to fi.nc-1 do
    begin
        buf[c*2  ]:=0;
        buf[c*2+1]:=-simg_large;
    end;
    d:=fi.pt;
    for i:=0 to fi.nx*fi.ny-1 do
    begin
        for c:=0 to fi.nc-1 do
        begin
            r:=d[c];
            buf[c*2]+=r;
            if r>buf[c*2+1] then buf[c*2+1]:=r;
        end;
        d+=fi.nc;
    end;
    i:=fi.nx*fi.ny;
    // calculate quantiles and threshold
    for c:=0 to fi.nc-1 do
    begin
        s:=buf[c*2]/i;
        buf[c*2]:=s+(SImg_GetQuantile(cc,fi,c,ref_quant)-s)*tf;
        buf[c*2+1]*=sat_fact;
//      writeln(stderr, buf[c*2]:0:3,'   ',buf[c*2+1]:0:3);
    end;
end;

{******* SImg_FWHMMap_findmax ********************************************}
{ thr     : threshold and saturation limit per color
  ni      : noise image (optional)
  rad     : minimum FWHM
  rad     : search radius
}
threadprocedure SImg_FWHMMap_findmax(fi:TSImg_FPImage; fk:TSImg_FPImage; ia:longint; iz:longint; thr:pdouble; dmin:double; rad:longint);
const k_fact = 2/pi;
      limit  = exp(ln(0.5)*4/(pi*k_fact));  { 0.25 @ k_fact=2/pi }
var d,e         : ^simgfp;
    r,s         : simgfp;
    a1,a2,a3    : double;
    rc1,rc2,rc3 : double;
    i,j,k,x,y,c : longint;
    l,kmin      : longint;
    xa,xz,xm    : longint;
    ya,yz,ym    : longint;
    sr1,sr2,sr3 : longint;
label l1;
begin
    xm:=0; ym:=0;   // makes the compiler happy
    kmin:=round(sqr(dmin)/k_fact);
    sr1:=sqr(rad-1);
    sr2:=round(sqr(rad*0.75));
    sr3:=round(sqr(rad*0.5));
    i:=0; j:=0; k:=0;
    for y:=-rad to rad do
        for x:=-rad to rad do
        begin
            l:=sqr(x)+sqr(y);
            if l>=sr1 then i+=1
            else if l>=sr2 then j+=1
            else if l>=sr3 then k+=1;
        end;
    rc1:=1/i;
    rc2:=1/j;
    rc3:=1/k;

    for i:=ia to iz do
    begin
        e:=fk.pt + i*fk.sl + fk.nc;
        ya:=i*fi.ny div fk.ny;
        yz:=(i+1)*fi.ny div fk.ny - 1;
        for j:=1 to fk.nx-2 do
        begin
            xa:=j*fi.nx div fk.nx;
            xz:=(j+1)*fi.nx div fk.nx - 1;
            for c:=0 to fi.nc-1 do
            begin
                { find maximum within square region of diameter rad }
                r:=-simg_large;
                for y:=ya to yz do
                begin
                    d:=fi.pt + y*fi.sl + xa*fi.nc + c;
                    for x:=xa to xz do
                    begin
                        s:=d^; d+=fi.nc;
                        if s>r then
                        begin
                            r:=s;
                            xm:=x;
                            ym:=y;
                        end;
                    end;
                end;
                if (r<thr[c*2]) or (r>thr[c*2+1]) then goto l1;
                { maximum within radius rad ?, background }
                a1:=0; a2:=0; a3:=0;
                for y:=-rad to rad do
                begin
                    d:=fi.pt + (ym+y)*fi.sl + (xm-rad)*fi.nc + c;
                    for x:=-rad to rad do
                    begin
                        s:=d^; d+=fi.nc;
                        if s>r then goto l1;
                        k:=sqr(x)+sqr(y);
                        if k>=sr1 then a1+=s
                        else if k>=sr2 then a2+=s
                        else if k>=sr3 then a3+=s;
                    end;
                end;
                a1*=rc1; a2*=rc2; a3*=rc3;
                if a2<a1 then a1:=a2;
                if a3<a1 then a1:=a3;
                { if a1<0 then a1:=0; }
                { calculate half maximum diameter }
                if r-a1<=thr[c*2] then goto l1;
                r:=a1*(1-limit) + r*limit;
                k:=0; l:=0;
                a2:=0;
                for y:=-rad to rad do
                begin
                    d:=fi.pt + (ym+y)*fi.sl + (xm-rad)*fi.nc + c;
                    for x:=-rad to rad do
                    begin
                        s:=d^; d+=fi.nc;
                        if s>r then
                        begin
                            k+=1;
                            l+=sqr(x)+sqr(y);
                            a2+=s;
                        end;
                    end;
                end;
                a2-=k*(a1+thr[c*2]);
                if (k<kmin) or (l>sqr(k-1)) or (a2<0) then goto l1;
                e[0]:=sqrt(k*k_fact);
                e[1]:=sqrt(a2);
l1:
                e+=2;
            end;
        end;
    end;
end;


{******* SImg_FWHMMap_int ************************************************}
{ fi      : input
  fj      : output
  dmin    : minimum FWHM
  rad     : search radius. Recommended 1.5..20 times the PSF width
}
procedure SImg_FWHMMap_int(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; dmin:double; rad:longint);
var fk          : TSImg_FPImage;
    d           : ^simgfp;
    r,s,t,u     : double;
    i,j,k,tn    : longint;
    buf,hist    : ^double;
threadvars;
begin
    if rad<5 then rad:=5;

    { thresholds }
    getmem2(buf,fi.nc*sizeof(double)*2);
    SImg_SharpnessFilter_threshold(cc,fi,buf,1);

    { map }
    fk.pt:=nil;
    SImg_New(cc, fk, fi.nx div rad, fi.ny div rad, fi.nc*2 );
    j:=1;
    tn:=cc.threads;
    for i:=0 to tn-1 do
    begin
        k:=(fk.ny-2)*(i+1) div tn + 1;
        dothread SImg_FWHMMap_findmax(fi,fk, j,k-1, buf, dmin, rad);
        j:=k;
    end;
    threadwait;

    { histogram }
    getmem2(hist,4*rad*sizeof(double));
    for j:=0 to fi.nc-1 do
    begin
        s:=1e-100;
        for i:=0 to 4*rad-1 do
            hist[i]:=0;
        { histogram }
        d:=fk.pt + j*2;
        for i:=0 to fk.nx*fk.ny-1 do
        begin
            r:=d[1];
            k:=round(d^*4)-2;
            if (k>=0) and (k<4*rad) then hist[k]+=r;
            s+=r;
            d+=fk.nc;
        end;
        { median u }
        t:=0; s*=0.5; u:=0;
        for i:=0 to 4*rad-1 do
        begin
            r:=hist[i];
            if t+r<s then u:=i
            else if (t<s) and (r>1e-50) then u:=i-1+(s-t)/r;
            t+=r;
{           if r>s*1e-6 then writeln(stderr,(i+2)*0.25:0:2,'   ',r/(2*s):0:6,'  ',j); }
        end;
{       writeln(stderr, 'median: ',(u+2)*0.25:0:4); }
        { drop values smaller than 0.5*u and larger than 2*u; multiply weight }
        u:=(u+2)*0.25*0.5;
        d:=fk.pt + j*2;
        for i:=0 to fk.nx*fk.ny-1 do
        begin
            r:=d^;
            if (r>=u) and (r<=4*u) then d^:=r*d[1]
            else begin
                d[0]:=0;
                d[1]:=0;
            end;
            d+=fk.nc;
        end;
    end;
    SImg_Move(cc,fk,fj);

    freemem2(hist,4*rad*sizeof(double));
    freemem2(buf,fi.nc*sizeof(double)*2);
end;


{*************************************************************************}
{******* SImg_FWHMMap ****************************************************}
{*************************************************************************}
{ Calculate star FWHM (in pixels) map
  fi      : input
  fj      : output
  dmin    : minimum FWHM
  rad     : search radius. Recommended 1.5..20 times the PSF width
  res     : weight smoothing relative to image dimension. Recommended: 2..10
  result  : total FWHM in pixels
}
function SImg_FWHMMap(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; dmin:double; rad:longint; res:double):double;
const reg  = 0.2;         { regularization using average }
var fk      : TSImg_FPImage;
    d,e     : ^simgfp;
    r,s,t   : double;
    i,j     : longint;
    buf     : ^double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_FWHMMap: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj)+' dmin='+real2str(dmin)+' rad='+int2str(rad)+' res='+real2str(res));
    if not SImg_Recover(cc,fi) then cc.error('SImg_FWHMMap: Invalid image: fi');
    if not SImg_OverwriteAlert(fi,fj) then SImg_Free(cc,fj);
    if rad<5 then rad:=5;

    { blurring radius }
    if res<1 then res:=1;
    i:=fi.nx;
    if fi.ny<i then i:=fi.ny;
    res:=i/(rad*res);
    if res<1.5 then res:=1.5;

    { get star widths }
    fk.pt:=nil;
    SImg_FWHMMap_int(cc,fi,fk,dmin,rad);

    { regularization coefficients }
    getmem2(buf,fk.nc*sizeof(double));
    for i:=0 to fk.nc-1 do
        buf[i]:=0;
    d:=fk.pt;
    for i:=0 to fk.nx*fk.ny-1 do
    begin
        for j:=0 to fk.nc-1 do
            buf[j]+=d[j];
        d+=fk.nc;
    end;

    t:=0;
    for i:=0 to fi.nc-1 do
    begin
        r:=reg*buf[i*2+1]/(fk.nx*fk.ny);
        s:=(buf[i*2]+1e-48)/(buf[i*2+1]+1e-50);
        buf[i*2  ]:=r*s;
        buf[i*2+1]:=r;
        t+=sqr(s);
    end;
    result:=sqrt(t/fi.nc);

    { blurring }
    SImg_Filter(cc,fk,fk,0.5*res);

    { output }
    SImg_New(cc, fj, fk.nx,fk.ny,fi.nc);
    d:=fk.pt; e:=fj.pt;
    for i:=0 to fk.nx*fk.ny-1 do
    begin
        for j:=0 to fj.nc-1 do
            e[j]:=(d[j*2]+buf[j*2]+1e-48)/(d[j*2+1]+buf[j*2+1]+1e-50);
        d+=fk.nc;
        e+=fj.nc;
    end;

    SImg_Free(cc,fk);
    freemem2(buf,fi.nc*sizeof(double)*2);
    SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_GetSharpnessFWHMM ******************************************}
{*************************************************************************}
{ Calculate star FWHM in pixels
  fi      : input
  dmin    : minimum FWHM
  rad     : search radius. Recommended 20*dmin
  fwhm    : array to store the FWHM per color. >rad if undefined
  result  : total FWHM in pixels
}
function SImg_GetSharpnessFWHM(cc:CSImg_Config; fi:TSImg_FPImage; dmin:double; rad:longint; var fwhm:array of double):double;
var fk      : TSImg_FPImage;
    d       : ^simgfp;
    s,t     : double;
    i,j     : longint;
    buf     : ^double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_SharpessFWHM: fi='+SImg_InfoStr2(fi)+' dmin='+real2str(dmin)+' rad='+int2str(rad));
    if not SImg_Recover(cc,fi) then cc.error('SImg_SharpnessFWHM: Invalid image: fi');
    if rad<5 then rad:=5;

    { get star widths }
    fk.pt:=nil;
    SImg_FWHMMap_int(cc,fi,fk,dmin,rad);

    { per color FWHM, total FWHM }
    getmem2(buf,fk.nc*sizeof(double));
    for i:=0 to fk.nc-1 do
        buf[i]:=0;
    d:=fk.pt;
    for i:=0 to fk.nx*fk.ny-1 do
    begin
        for j:=0 to fk.nc-1 do
            buf[j]+=d[j];
        d+=fk.nc;
    end;

    t:=0;
    for i:=0 to fi.nc-1 do
    begin
        s:=(buf[i*2]+1e-30)/(buf[i*2+1]+2*rad*1e-30);
        if i<=high(fwhm) then fwhm[i]:=s;
        t+=sqr(s);
    end;
    result:=sqrt(t/fi.nc);

    SImg_Free(cc,fk);
    freemem2(buf,fi.nc*sizeof(double)*2);
    SImg_DecDebugLevel;
end;

function SImg_GetSharpnessFWHM(cc:CSImg_Config; fi:TSImg_FPImage; dmin:double; rad:longint):double;
var a : array[0..0] of double;
begin
    result:=SImg_GetSharpnessFWHM(cc,fi,dmin,rad,a);
end;


function SImg_GetSharpnessFWHM(cc:CSImg_Config; fi:TSImg_FPImage):double;
begin
    result:=SImg_GetSharpnessFWHM(cc,fi, 1.25,25);
end;


{*************************************************************************}
{******* SImg_GetSingleSource ********************************************}
{*************************************************************************}
{ Extract a single source from a 1 color image. This is intended for measurements with artificial stars.
    fi     : input image
    fs     : PSF output image
    rad    : Radius of output image and background subtraction radius. This is also the minimum distance from border.
    result : Background which is subtracted from input image
}
function SImg_GetSingleSource(cc:CSimg_Config; fi:TSImg_FPImage; var fs:TSImg_FPImage; rad:longint):double;
const bl_quant = 0.1;   { black level quantile }
      bp_thr   = 10;    { bad pixel threshold, see below }
var fj,fk               : TSImg_FPImage;
    x,y,xc,yc,i,j       : longint;
    a,b                 : simgfp;
    r                   : double;
    d,e                 : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_Default;
{    SImg_IncDebugLevel;
    cc.debug('SImg_GetSingleSource: fi='+SImg_InfoStr2(fi)+' fs='+SImg_InfoStr2(fs)+' rad='+int2str(rad)); }
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetSingleSource: Invalid image: fi');
    if fi.nc<>1 then cc.error('SImg_GetSingleSource: One color image expected.');
    if not SImg_OverwriteAlert(fi,fs) then SImg_Free(cc,fs);
    if rad<3 then cc.error('SImg_GetSingleSource: Radius to small');
    x:=2*rad+1;
    if (fi.ny<x) or (fi.nx<x) then cc.error('SImg_GetSingleSource: Dimension of input image to small or radius to large.');

    { defect pixel filter: pixels must not be brighter than bp_thr times the average or the 8 surrounding pixels }
    fj.pt:=nil;
    SImg_New(cc, fj, fi.nx,fi.ny,1, false);
    b:=SImg_GetQuantile(cc, fi, bl_quant);
    for y:=0 to fi.ny-1 do
    begin
        d:=fi.pt + y*fi.sl;
        e:=fj.pt + y*fj.sl;
        if (y=0) or (y=fi.ny-1) then
//      if true then
        begin
            for x:=0 to fi.nx-1 do
                e[x]:=d[x];
        end else
        begin
            e[0]:=d^; d+=1;
            for x:=1 to fi.nx-2 do
            begin
                r:=(  d[-fi.sl-1] + d[-fi.sl] + d[-fi.sl+1]
                    + d[      -1]             + d[       1]
                    + d[ fi.sl-1] + d[ fi.sl] + d[ fi.sl+1] )*0.125;
                a:=d^;
                if a-b>(r-b)*bp_thr then e[x]:=r
                else e[x]:=a;
                d+=1;
            end;
            e[fi.nx-1]:=d^;
        end;
    end;
//    SImg_WriteImageFile(cc,'getSingelSource-fj.fits',fj);

    { background filter }
    fk.pt:=nil;
    SImg_Filter(cc, fj,fk, rad);

    { find center of source }
    xc:=-1; yc:=-1;
    a:=simg_mina;
    for y:=rad to fi.ny-rad-1 do
    begin
        d:=fj.pt + y*fj.sl;
        e:=fk.pt + y*fk.sl;
        for x:=rad to fi.nx-rad-1 do
        begin
            b:=d[x]-e[x];
            if b>a then
            begin
                xc:=x;
                yc:=y;
                a:=b;
            end;
        end;
    end;
    if xc<0 then cc.error('SImg_GetSingleSource: No peak found, image seem to be constant');
    SImg_Free(cc, fk);
    cc.info(cc.name+': Found single source at ('+int2str(xc)+', '+int2str(yc)+')');

    { extract PSF and subtract background }
    SImg_New(cc, fs,  2*rad+1, 2*rad+1, 1);
    i:=0; r:=0;
    j:=round(sqr(rad*0.95));
    for y:=-rad to rad do
    begin
        d:=fj.pt + (y+yc)*fj.sl + xc;
        e:=fs.pt + (y+rad)*fs.sl + rad;
        for x:=-rad to rad do
        begin
            a:=d[x];
            e[x]:=a;
            if sqr(x)+sqr(y)>=j then
            begin
                r+=a;
                i+=1;
            end;
        end;
    end;
    SImg_Free(cc, fj);
    r/=i; b:=r;
    e:=fs.pt;
    for i:=0 to fs.size-1 do
        e[i]-=b;
//    SImg_WriteImageFile(cc,'getSingelSource-fs.fits',fs);

    result:=r;
{    SImg_DecDebugLevel; }
end;


{*************************************************************************}
{******* SImg_SingleSourceFWHM *******************************************}
{*************************************************************************}
{ Extract a single source and calulate its FWHM. This is intended for measurements with artificial stars.
    fi     : input image
    fs     : PSF output image
    rad    : Radius of output image and background subtraction radius. This is also the minimum distance from border.
    fwhm   : array to store the FWHM per color.
    bg     : array to store the background per color
    result : RMS FWHM
}
function SImg_SingleSourceFWHM(cc:CSImg_Config; fi:TSImg_FPImage; var fs:TSImg_FPImage; rad:longint; var fwhm,bg:array of double):double;
var fj,fk,fo    : TSImg_FPImage;
    i,c         : longint;
    r,s         : double;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_SingleSourceFWHM: fi='+SImg_InfoStr2(fi)+' fs='+SImg_InfoStr2(fs)+' rad='+int2str(rad));
    if not SImg_Recover(cc,fi) then cc.error('SImg_GetSingleSource: Invalid image: fi');
    if not SImg_OverwriteAlert(fi,fs) then SImg_Free(cc,fs);
    if rad<3 then cc.error('SImg_SingleSourceFWHM: Radius to small');

    if fi.nc=1 then fj:=fi
    else begin
        fj.pt:=nil;
        SImg_New(cc, fj, fi.nx,fi.ny,1, false);
    end;

    s:=0;
    fo.pt:=nil; fk.pt:=nil;
    SImg_New(cc, fo, 2*rad+1, 2*rad+1, fi.nc);
    for c:=0 to fi.nc-1 do
    begin
        if fi.pt<>fj.pt then
            for i:=0 to fj.size-1 do
                fj.pt[i]:=fi.pt[i*fi.nc+c];
        r:=SImg_GetSingleSource(cc, fj,fk, rad);
        if c<=high(bg) then bg[c]:=r;
//    writeln(r);
        r:=SImg_GetPSFFWHM(cc, fk);
        if c<=high(fwhm) then fwhm[c]:=r;
        s+=sqr(r);
        for i:=0 to fk.size-1 do
            fo.pt[i*fo.nc+c]:=fk.pt[i];
    end;
    SImg_Free(cc, fk);
    if fj.pt<>fi.pt then SImg_Free(cc, fj);
    SImg_Move(cc, fo,fs);

    result:=sqrt(s/fi.nc);
    SImg_DecDebugLevel;
end;

{ Extract a single source and calulate its FWHM. This is intended for measurements with artificial stars.
    fi     : input image
    rad    : Radius of output image and background subtraction radius. This is also the minimum distance from border.
    fwhm   : RMS FWHM
    result : brightest pixel
}
function SImg_SingleSourceFWHM(cc:CSImg_Config; fi:TSImg_FPImage; rad:longint; out fwhm:double):simgfp;
var fw,bg       : array of double;
    fs          : TSImg_FPImage;
    a,b         : simgfp;
    c,i         : longint;
begin
    setlength(fw,0);
    setlength(bg,max(0,fi.nc));
    fs.pt:=nil;
    fwhm:=SImg_SingleSourceFWHM(cc, fi,fs, rad, fw,bg);
    b:=-simg_large;
    for i:=0 to fs.nx*fs.ny-1 do
        for c:=0 to fs.nc-1 do
        begin
            a:=fs.pt[i*fs.nc+c]+bg[c];
            if a>b then b:=a;
        end;
    result:=b;
    SImg_Free(cc,fs);
    setlength(bg,0);
end;
